from base_aux.testplans.tc import *
from base_aux.valid.m2_valid_base2_derivatives import *
from base_aux.valid.m3_valid_chains import *

from . import tc1_direct
from .tc0_groups import *


# =====================================================================================================================
class TestCase(TcGroup_ATC220220, tc1_direct.TestCase):
    ASYNC = True
    DESCRIPTION = "reverse1"

    @classmethod
    def startup__cls__wrapped(cls) -> TYPE__RESULT_W_NORETURN:
        result_chain = ValidChains(
            [
                Valid(value_link=True, name="TRUE"),
                Valid(value_link=False, name="FALSE"),
                Valid(value_link=None, name="NONE"),
            ],
        )
        return result_chain

    def run__wrapped(self) -> bool:
        time.sleep(0.1)
        self.details_update({"detail_value": not self.DEVICES__BREEDER_INST.DUT.VALUE})
        return not self.DEVICES__BREEDER_INST.DUT.VALUE


# =====================================================================================================================



from typing import *

from base_aux.base_statics.m2_exceptions import *


# =====================================================================================================================
class NoValue:
    """
    TODO: DEPRECATE???=NO! used in valid/value
    ---------
    use direct ArgsEmpty???/ArgsKwargs()

    NOTE
    ----
    never instantiate it! use value only as Class!

    GOAL
    ----
    it is different from Default!
    there is no value!
    used when we need to change logic with not passed value!

    SPECIALLY CREATED FOR
    ---------------------
    Valid as universal validation object under cmp other aux_types!

    USAGE
    -----
    class Cls:
        def __init__(self, value: Any | type[NoValue] | NoValue = NoValue):
            self.value = value

        def __eq__(self, other):
            if self.value is NoValue:
                return other is True
                # or
                return self.__class__(other).run()
            else:
                return other == self.value

        def run(self):
            return bool(self.value)
    """
    def __init__(self) -> NoReturn:
        raise Exx__WrongUsage(NoValue)

    def __bool__(self):
        return False

    # todo: add classmethod! - not working!!!
    def __str__(self):
        return f"{self.__class__.__name__}"


# =====================================================================================================================



from typing import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .m0_base import Data_


# =====================================================================================================================
class HeaderViewCB(QHeaderView):
    """
    THIS IS THE ONLY WAY TO PUT CHECKBOXES INTO HEADER!

    work as additional to tableModel!
    this is always exists in TV as default! but we can change some behaviour!
    """
    # SIGNALS ---------------------------------------------
    signal__state_changed = pyqtSignal(list)
    sectionCountChanged: pyqtSignal
    sectionClicked: pyqtSignal
    sectionDoubleClicked: pyqtSignal
    sectionPressed: pyqtSignal
    sectionEntered: pyqtSignal
    sectionMoved: pyqtSignal
    geometriesChanged: pyqtSignal

    # ORIGINAL ---------------------------------------------
    setSectionHidden: Callable  # (logicalIndex, hide)
    setSectionsMovable: Callable  # (movable)
    setVisible: Callable  # (visible)
    resizeSection: Callable  # (logicalIndex, size)

    # NEW --------------------------------------------------
    DATA: Data_ = None

    def __init__(self, data: Data_, orientation=Qt.Horizontal, parent=None):
        self.DATA = data
        super().__init__(orientation, parent)
        self.setSectionsClickable(True)
        self.sectionClicked.connect(self.on_sectionClicked)

        # self.setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        # self.setSectionResizeMode(0, QHeaderView.ResizeMode.ResizeToContents)
        # self.setSectionResizeMode(2, QHeaderView.ResizeMode.ResizeToContents)

    def count(self):
        return len(self.DATA.DEVS)

    # def mouseDoubleClickEvent(self, e, QMouseEvent=None): # real signature unknown; restored from __doc__
    #     """ mouseDoubleClickEvent(self, e: Optional[QMouseEvent]) """
    #     pass

    # def mousePressEvent(self, e, QMouseEvent=None):
    #     """ mousePressEvent(self, e: Optional[QMouseEvent]) """
    #     pass

    def paintSection(self, painter: Optional[QPainter], rect: QRect, logicalIndex: int) -> None:
        painter.save()
        QHeaderView.paintSection(self, painter, rect, logicalIndex)
        painter.restore()

        if logicalIndex > 0:
            option = QStyleOptionButton()
            # option.initFrom(self)             # not necessary
            # option.iconSize = QSize(10, 10)   # not necessary

            # var1=чекбокс позиционирование -------------------------
            # option.rect = rect  # center
            # option.rect = QRect(10, 10, 10, 10)
            # option.rect = QRect(rect.left(), 0, 20, 20)
            option.rect = QRect(rect.left(), rect.top(), 20, 20)

            dut = self.DATA.DEVS[logicalIndex - 1]
            if not dut.SKIP:
                option.state |= QStyle.State_On
            else:
                option.state |= QStyle.State_Off
            self.style().drawPrimitive(QStyle.PE_IndicatorCheckBox, option, painter)

    def on_sectionClicked(self, index: int) -> None:
        if index > 0:
            dut = self.DATA.DEVS[index - 1]
            dut.SKIP_reverse()
            # self.TM._data_reread()    # DONT NEED!!!!


# =====================================================================================================================



from typing import *

from base_aux.aux_attr.m2_annot1_aux import *
from base_aux.breeders.m1_breeder_str1_series import *
from base_aux.aux_attr.m1_attr2_nest1_gsai_anycase import *


# =====================================================================================================================
class BreederStrStack(NestGAI_AttrIC):
    """
    CREATED SPECIALLY FOR
    ---------------------
    applying in Gui tableModels (PyQt5) as header structure

    VULNERABILITIES # FIXME:
    ------------------------
    1. if exists same names cause of patterns - it would return always first index!
        class BreederStrStack_Example2(BreederStrStack):
            TAIL = BreederStrSeries(2, 2, "%s")
            TAIL2 = BreederStrSeries(4, 2, "%s")

        BreederStrStack_Example2()["2"] = 1

    DEFINE
    ------
    !. names (as attributes)
        - dont use underscore as first simble
        - it will be as VALUE for index
    1. dont keep SKIPPED indexes - use all final range!
    2. BreederStrSeries
     - use if need template for some range
     - use any count of such items

    USAGE
    -----
    1. COMPARE BY INDEXES OVER NAME
    if colIndex == MyHeders.attr0:
        pass

    2. GET HEADER NAME BY INDEX or INDEX by NAME!
    headerName = MyHeders[colIndex]
    headerIndex = MyHeders[colName]

    RULES
    -----
    1. use always any ANNOTATIONS for your indexes!!
    2. hide names by using underscore!
    3. use None for AUTOINDEX!!!
    4. indexes follow the order
        class BreederStrStack_Example(BreederStrStack):
            atr1: int
            atr2: int = None


    5. nesting available with correct order!
        class ClsFirst(BreederStrStack):
            atr1: int
            atr3: int = None

        class ClsLast(BreederStrStack):
            atr2: int = None
            atr4: int

        for key, VALUE in ClsLast.annotations__get_nested_list().items():
            print(f"{key}:{VALUE}")

        # atr1:<class 'int'>
        # atr3:<class 'int'>
        # atr2:<class 'int'>
        # atr4:<class 'int'>
    """
    # settings ----------------------
    _INDEX_START: int | None = 0
    _RAISE_IF_INDEX_SKIPPED: bool = True

    # aux ----------------------
    _DATA: dict[int, str] = {}

    def __init__(self):
        if self._INDEX_START is None:
            self._INDEX_START = 0

        index_last = self._INDEX_START -1
        result = {}

        # ATTRS ------------------
        attrs: list[str] = []
        for attr in AnnotsAux(self).get__dict_values():
            if not attr.startswith("_") and not callable(getattr(self, attr)):
                attrs.append(attr)

        # WORK -------------------
        for attr in attrs:
            index = getattr(self, attr)

            # apply AUTO ----------
            if index is None:
                index = index_last + 1
                index_last = index
                setattr(self, attr, index)

            # work INT ----------
            if isinstance(index, int):
                if index in result:
                    msg = f"{index=} from {result=}"
                    raise Exx__IndexOverlayed(msg)
                result.update({index: attr})
                index_last = index

            # work BreederStrSeries ----------
            if isinstance(index, BreederStrSeries):
                # apply AUTO by recreation
                if index.START_OUTER is None:
                    index = BreederStrSeries(index_last + 1, index.COUNT, index.TEMPLATE, index.START_INNER)
                    setattr(self, attr, index)

                # work
                result_sub_dict = index.get_dict__outer()
                for key, value in result_sub_dict.items():
                    if key in result:
                        msg = f"{key=} from {result_sub_dict=}"
                        raise Exx__IndexOverlayed(msg)
                    result.update({key: value})
                    index_last = key

        # RESULT -------------------
        self._DATA = dict(sorted(result.items()))

        # CHECK SKIPPED -----------
        if self._RAISE_IF_INDEX_SKIPPED:
            self.raise_if_index_skipped()

    def __getitem__(self, item: Union[int, str]) -> Union[int, str]:
        if item in self._DATA:
            return self._DATA[item]

        if item in self._DATA.values():
            for name, value in self._DATA.items():
                if item == value:
                    return name

        msg = f"{item=}"
        raise Exx__NotExists(msg)

    def __contains__(self, item: Union[int, str]) -> bool:
        """
        :param item: OUTER INDEX (not self INNER!!!) or inner VALUE NAME
        :return:
        """
        return item in self._DATA or item in self._DATA.values()

    def count(self) -> int:
        return len(self._DATA)

    @classmethod
    def raise_if_index_skipped(cls) -> None | NoReturn:
        index_prev = None
        for index in cls._DATA:
            if index_prev is None:
                index_prev = index
                continue

            if index - index_prev != 1:
                msg = f"index [{index-1}]"
                raise Exx__IndexNotSet(msg)
            index_prev = index


# =====================================================================================================================
class BreederStrStack_Example(BreederStrStack):
    name0: int = 0
    name1: int = 1
    TAIL: BreederStrSeries = BreederStrSeries(2, 2, "%s")


class BreederStrStack_Example__BestUsage(BreederStrStack):
    INDEX: int = 0
    TESTCASE: int = None
    ASYNC: int = None
    STARTUP: int = None
    DUTS: BreederStrSeries = BreederStrSeries(None, 10, "TC_DUT_%s")


# =====================================================================================================================



from typing import *
from base_aux.aux_values.m0_novalue import *


# =====================================================================================================================
class NestInit_Source:
    """
    GOAL
    ----
    just show that class uses init with source - one param
    means inside instance all methods will work on it!
    and all params for methods belong to methods! not about new source!

    SPECIALLY CREATED FOR
    ---------------------
    apply in AttrAux and same others

    :ivar SOURCE: use Lambda or even simple Callable to make a callableGenerate default value! like dict or other user class.
        main Idea for Callable Source is generate independent value in instances.
        it keeps callable type only in class attribute! in instance it will be resolved by calling!

    BEST USAGE
    ----------
        class ClsAux(NestInit_Source):
            SOURCE = MyClass
            SOURCE = Lambda(dict)
            SOURCE = dict
    """
    # SOURCE: dict = Lambda(dict)               # for callable
    # SOURCE: Any | Lambda = Lambda(None)       # generic final value
    SOURCE: Any = None                          # generic final value
    # SOURCE_DEF__CALL_IF_CALLABLE: bool = True   # used only for CLS.SOURCE! not for param source! really it is NOT NEED!

    @classmethod
    @property
    def SOURCE_DEF(cls) -> Any | NoReturn:
        # if isinstance(cls.SOURCE, Lambda):
        if callable(cls.SOURCE):        # and cls.SOURCE_DEF__CALL_IF_CALLABLE:
            result = cls.SOURCE()
        else:
            result = cls.SOURCE
        return result

    def __init__(self, source: Any = NoValue, *args, **kwargs) -> None | NoReturn:
        super().__init__(*args, **kwargs)
        self.init_source(source)
        self.init_post()

    def init_source(self, source: Any = None) -> None | NoReturn:
        if source is not NoValue:
            self.SOURCE = source
        else:
            self.SOURCE = self.SOURCE_DEF

    def init_post(self) -> None | NoReturn:
        """
        GOAL
        ----
        user initions

        TYPICAL USAGE
        -------------
        make some conversations for source, like str for text
        or
        make initial tests/checks for source, like typecheck
        """
        return NotImplemented


# =====================================================================================================================



from typing import *
from base_aux.aux_dict.m3_dict_attr1_simple import *


# =====================================================================================================================
class DictAttrFull(DictAttr):
    """
    TODO: do this perfect idea!

    smth like BS4 syntax but for dict/list
    1. use full access WR work with any level
    2. dict/list - ok! tuple - as generic final!
    3. keep source link! not just a simple copy (like in DictAttr)!

    WHY NOT - DictAttr
    ------------------
    it was a simple way to work with mainly first level! and only with dict!
    it seems it is obviously need use with lists same!
    """


# =====================================================================================================================



from typing import *

from base_aux.base_inits.m1_nest_init_source import *
from base_aux.base_statics.m4_enums import *


# =====================================================================================================================
@final
class DictAux(NestInit_Source):
    """
    NOTE
    ----
    decide where to work - source or copy????
    """
    SOURCE: dict[Any, Any] = dict

    # -----------------------------------------------------------------------------------------------------------------
    def clear_values(self, form: FormIntExt = FormIntExt.EXTERNAL) -> dict[Any, None]:
        keys = list(self.SOURCE)
        new_dict = dict.fromkeys(keys)
        if form == FormIntExt.EXTERNAL:
            return new_dict

        if form == FormIntExt.INTERNAL:
            self.SOURCE.clear()
            self.SOURCE.update(new_dict)
            return self.SOURCE

    # -----------------------------------------------------------------------------------------------------------------
    def keys_del(self, *keys: Any) -> None:
        for key in keys:
            try:
                self.SOURCE.pop(key)
            except:
                pass

    def keys_rename__by_func(self, func: Callable[[Any], Any], form: FormIntExt = FormIntExt.EXTERNAL) -> dict[Any, Any]:
        """
        GOAL
        ----
        useful to rename keys by func like str.LOWER/upper
        raise on func - delete key from origin! applied like filter
        """
        result = {}
        if form == FormIntExt.EXTERNAL:
            result = {}
        elif form == FormIntExt.INTERNAL:
            result = self.SOURCE

        # -----------------------
        for key in list(self.SOURCE):
            value = self.SOURCE.get(key)
            DictAux(result).keys_del(key)
            try:
                key_new = func(key)
                result.update({key_new: value})
            except:
                pass

        return result

    # -----------------------------------------------------------------------------------------------------------------
    def collapse_key(self, key: Any) -> dict[Any, Any]:
        """
        GOAL
        ----
        specially created for 2level-dicts (when values could be a dict)
        so it would replace values (if they are dicts and have special_key)

        CONSTRAINTS
        -----------
        it means that you have similar dicts with same exact keys
            {
                0: 0,
                1: {1:1, 2:2, 3:3},
                2: {1:11, 2:22, 3:33},
                3: {1:111, 2:222, 3:333},
                4: 4,
            }
        and want to get special slice like result

        SPECIALLY CREATED FOR
        ---------------------
        testplans get results for special dut from all results


        main idia to use values like dicts as variety and we can select now exact composition! remain other values without variants

        EXAMPLES
        --------
        dicts like
            {
                1: {1:1, 2:2, 3:3},
                2: {1:1, 2:None},
                3: {1:1},
                4: 4,
            }
        for key=2 return
            {
                1: 2,
                2: None,
                3: None,
                4: 4,
            }

        """
        result = {}
        for root_key, root_value in self.SOURCE.items():
            if isinstance(root_value, dict) and key in root_value:
                root_value = root_value.get(key)

            result[root_key] = root_value

        return result

    # -----------------------------------------------------------------------------------------------------------------
    def prepare_serialisation(self) -> dict:
        result = {}
        # TODO: FINISH

        return result


# =====================================================================================================================



from base_aux.aux_attr.m2_annot1_aux import *
from base_aux.aux_attr.m1_attr2_nest1_gsai_anycase import *
from base_aux.aux_cmp_eq.m2_eq_aux import *


# =====================================================================================================================
class NestInit_AnnotsAttrByKwArgs(NestGAI_AttrIC):     # NOTE: dont create AnnotsOnly/AttrsOnly! always use this class!
    """
    NOTE
    ----
    for more understanding application/logic use annots at first place! and dont mess them. keep your code clear!
        class Cls(NestInit_AnnotsAttrByKwArgs):
            A1: Any
            A2: Any
            A3: Any = 1
            A4: Any = 1

    GOAL
    ----
    init annots/attrs by params in __init__

    LOGIC
    -----
    ARGS
        - used for ANNOTS ONLY - used as values! not names!
        - inited first without Kwargs sense
        - if args less then annots - no matter
        - if args more then annots - no matter+no exx
        - if kwargs use same keys - it will overwrite by kwargs (args set first)
    KWARGS
        - used for both annots/attrs (annots see first)
        - if not existed in Annots and Attrs - create new!
    """
    def __init__(self, *args: Any, **kwargs: TYPING.KWARGS_FINAL) -> None | NoReturn:
        AnnotsAux(self).set_annots_attrs__by_args_kwargs(*args, **kwargs)
        AnnotsAux(self).check_all_defined_or_raise()    # fixme: is it really need? i think yes! use default values for noRaise!


# ---------------------------------------------------------------------------------------------------------------------
# class NestInit_AnnotsAttrByKwArgsIc(NestInit_AnnotsAttrByKwArgs, NestGSAI_AttrAnycase):   # IC - IS NOT WORKING!!!
#     """
#     SAME AS - 1=parent
#     -------
#     but attrs access will be IgnoreCased
#     """
#     pass


# ---------------------------------------------------------------------------------------------------------------------
def examples__NestInit():
    class Example(NestInit_AnnotsAttrByKwArgs):
        A1: Any
        A2: Any = None
        A3 = None

    try:
        Example()
        assert False
    except:
        assert True

    assert Example(a1=1).A1 == 1
    assert Example(1, a1=2).A1 == 2

    assert Example(1).A1 == 1
    assert Example(1).A2 == None
    assert Example(1).A3 == None

    assert Example(1, 1, 1).A1 == 1
    assert Example(1, 1, 1).A2 == 1
    assert Example(1, 1, 1).A3 == None
    assert Example(1, 1, a3=1).A3 == 1


# =====================================================================================================================
if __name__ == '__main__':
    pass


# =====================================================================================================================



from base_aux.aux_attr.m4_kits import AttrKit_Blank
from base_aux.path2_file.m3_filetext import *
from base_aux.base_inits.m3_nest_init_annots_attrs_by_kwargs import *


# =====================================================================================================================
class AttrsLoader_DictTextFile(TextFile, NestCall_Resolve):
    """
    GOAL
    ----
    load attrs as parsed dict
    main usage is final key values!
    used for get settings from file into NestInit_AnnotsAttrByKwArgsIc
    """
    TARGET: type[NestInit_AnnotsAttrByKwArgs] | Any = AttrKit_Blank
    STYLE: DictTextFormat = DictTextFormat.AUTO
    KEYPATH: tuple[str | int, ...] = ()

    FILEPATH: pathlib.Path
    TEXT: str

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(
            self,
            target: type | Any = None,
            keypath: tuple[str | int, ...] = None,     # path to exact dict in dict
            style: DictTextFormat = None,

            **kwargs,       # init File/Text
    ) -> None | NoReturn:
        super().__init__(**kwargs)

        self.init_style(style)

        if target is not None:
            self.TARGET = target

        if keypath is not None:
            self.KEYPATH = keypath

    def init_style(self, style: DictTextFormat) -> None:
        if style is not None:
            self.STYLE = DictTextFormat(style)

        if self.STYLE == DictTextFormat.EXTENTION:
            for item in DictTextFormat:
                if self.FILEPATH.name.lower().endswith(str(item.value).lower()):
                    self.STYLE = item
                    break

        if self.STYLE is None:
            pass

    # -----------------------------------------------------------------------------------------------------------------
    def resolve(self) -> NestInit_AnnotsAttrByKwArgs | Any | NoReturn:
        # get dict -------
        data = self.parse__dict(self.STYLE)
        if data is None:
            raise Exx__Incompatible(f"{self.STYLE=}/{self.TEXT=}")

        # load keypath ---
        if self.KEYPATH:
            data = IterAux(data).value__get(*self.KEYPATH)

        # load args -------
        if TypeAux(self.TARGET).check__class() and issubclass(self.TARGET, NestInit_AnnotsAttrByKwArgs):
            # used for check Annots all inited!

            result = self.TARGET(**data)
        else:
            AnnotsAux(self.TARGET).set_annots_attrs__by_args_kwargs(**data)
            result = self.TARGET

        return result


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



from typing import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from base_aux.testplans.tc import *

from base_aux.pyqt.m0_static import *
from base_aux.pyqt.m0_base1_hl import *
from base_aux.pyqt.m4_gui import *
from base_aux.pyqt.m2_mods import *

from .tm import TpTableModel
from .dialog import DialogsSetTp


# =====================================================================================================================
class TpHlStyles(HlStyles):
    RESULT: HlStyle = HlStyle(
        FORMAT=format_make("", "", "lightGrey"),
        P_ITEMS=[
            "validate_last_bool", "result",
        ],
        P_TEMPLATES=[
            r'.*%s.*',
        ],
    )
    RESULT_TRUE: HlStyle = HlStyle(
        FORMAT=format_make("", "", "lightGreen"),
        P_ITEMS=[
            "True",
        ],
        P_TEMPLATES=[
            r'.*validate_last_bool=%s.*',
            r'.*result[^=]*=%s.*',
            # r'.*valid[^=]*=%s.*',
        ],
    )
    RESULT_FALSE: HlStyle = HlStyle(
        FORMAT=format_make("", "", "pink"),
        P_ITEMS=[
            "False",
        ],
        P_TEMPLATES=[
            r'.*validate_last_bool=%s.*',
            r'.*result[^=]*=%s.*',
            # r'.*valid[^=]*=%s.*',
        ],
    )
    # yellow -----
    RESULT_FALSE_NOCUM: HlStyle = HlStyle(
        FORMAT=format_make("", "", "yellow"),
        P_ITEMS=[
            "ValidNoCum", "ValidSleep", "ValidBreak",
        ],
        P_TEMPLATES=[
            r'.*%s.*validate_last_bool=False.*',
        ],
    )
    SKIPPED_LINE: HlStyle = HlStyle(
        FORMAT=format_make("", "", "yellow"),
        P_ITEMS=[
        ],
        P_TEMPLATES=[
            r'.*skip_last=True.*',
        ],
    )
    BREAK: HlStyle = HlStyle(
        FORMAT=format_make("", "", "yellow"),
        P_ITEMS=[
            "ValidBreak",
        ],
        P_TEMPLATES=[
            r'.*%s.*validate_last_bool=True.*',
        ],
    )


PTE_RESULTS_EXAMPLE = """
ПРОВЕРКА выделения текста результатов
Valid=None    #без явного bool значения
Result=True   #явное True
Valid=True    #явное True
Valid=False   #явное False
ValidNoCum(123x=False, validate_last_bool=True)
ValidNoCum(123x=True, validate_last_bool=False)  #неважный/неучтенный False

result__startup=True
------------------------------------------------------------
result=Valid(NAME=GET,skip_last=False,validate_last_bool=True,
...VALUE_LINK=<function SerialClient.__getattr__.<locals>.<lambda> at 0x000001E5AABF6FC0>,ARGS__VALUE=('PRSNT',),value_last=0,
...VALIDATE_LINK=0,validate_last=True,
,finished=True,timestamp_last=1730375894.4305248,
------------------------------------------------------------
result__teardown=ValidChains(NAME=,skip_last=False,validate_last_bool=True,
...VALUE_LINK=None,value_last=None,
...VALIDATE_LINK=True,validate_last=True,
,finished=True,timestamp_last=1730375895.0524113,
result__startup=ValidSleep(NAME=Sleep,skip_last=False,validate_last_bool=True,
___0:(START) len=3/timestamp=1730375895.0524113
___1:ValidNoCum(NAME=DUT.connect__only_if_address_resolved,skip_last=False,validate_last_bool=False,
___1:ValidNoCum(NAME=DUT.connect__only_if_address_resolved,skip_last=False,validate_last_bool=True,
___1:ValidNoCum(NAME=DUT.connect__only_if_address_resolved,skip_last=False,validate_last_bool=None,
___1:Valid(NAME=DUT.connect__only_if_address_resolved,skip_last=False,validate_last_bool=True,
...VALUE_LINK=<bound method SerialClient.connect__only_if_address_resolved of <DEVICES.ptb.Device object at 0x000001E5AAB4B050>>,value_last=True,
...VALIDATE_LINK=True,validate_last=True,
,finished=True,timestamp_last=1730375895.0524113,
___2:ValidRetry1(NAME=RST,skip_last=False,validate_last_bool=True,
...VALUE_LINK=<function SerialClient.__getattr__.<locals>.<lambda> at 0x000001E5AABF7240>,value_last=OK,
...VALIDATE_LINK=OK,validate_last=True,
,finished=True,timestamp_last=1730375896.3252466,
___3:ValidSleep(NAME=Sleep,skip_last=False,validate_last_bool=True,
...VALUE_LINK=<built-in function sleep>,ARGS__VALUE=(1,),value_last=None,
...VALIDATE_LINK=None,validate_last=True,
,finished=True,timestamp_last=1730375896.9521606,
___4:(FINISH) [result=False]/len=3
------------------------------------------------------------
DETAILS=====================
"""


# =====================================================================================================================
class TpGuiBase(Gui):
    # OVERWRITTEN -----------------------------------
    START = False

    TITLE = "[TestPlan] Title"
    SIZE = (1500, 800)

    HL_STYLES = TpHlStyles()
    DIALOGS = DialogsSetTp

    # NEW -------------------------------------------
    DATA: "TpMultyDutBase"

    def __init__(self, data):
        self.TITLE = f"[Тестплан]{data.STAND_NAME}/{data.STAND_DESCRIPTION[:20]}"
        super().__init__(data)

    def main_window__finalise(self):
        super().main_window__finalise()
        self.BTN_extended_mode__toggled()

    # WINDOW ==========================================================================================================
    def wgt_create(self):
        self.BTN_create()
        self.CB_create()
        self.TV_create()
        self.PTE_create()
        self.HL_create()

        # LAYOUT_CONTROL ----------------------------------------------------------------------------------------------
        layout_control = QVBoxLayout()
        layout_control.setAlignment(ALIGNMENT.T)

        layout_control.addWidget(self.BTN_devs_detect)
        layout_control.addWidget(self.BTN_start)
        layout_control.addWidget(self.CB_tp_run_infinit)
        layout_control.addWidget(self.CB_tc_run_single)
        layout_control.addWidget(self.BTN_save)
        layout_control.addStretch()

        layout_control.addWidget(self.BTN_settings)
        layout_control.addWidget(self.BTN_clear_all)
        layout_control.addWidget(self.BTN_reset_all)
        layout_control.addWidget(self.BTN_extended_mode)

        # LAYOUT_MAIN -------------------------------------------------------------------------------------------------
        self.LAYOUT_MAIN = QHBoxLayout()
        self.LAYOUT_MAIN.addLayout(layout_control)
        self.LAYOUT_MAIN.addWidget(self.TV)
        self.LAYOUT_MAIN.addWidget(self.PTE)

        # CENTRAL -----------------------------------------------------------------------------------------------------
        self.CENTRAL_WGT.setLayout(self.LAYOUT_MAIN)

    # WGTS ============================================================================================================
    def BTN_create(self) -> None:
        self.BTN_devs_detect = QPushButton("определить устройства")

        self.BTN_start = QPushButton_Checkable(["ТЕСТИРОВАНИЕ запустить", "ТЕСТИРОВАНИЕ остановить"])
        self.BTN_start.setCheckable(True)

        self.BTN_settings = QPushButton("настройки")
        self.BTN_settings.setCheckable(True)

        self.BTN_save = QPushButton("Сохранить результаты")
        self.BTN_clear_all = QPushButton("Очистить все результаты")
        self.BTN_reset_all = QPushButton("Отключить все устройства")

        self.BTN_extended_mode = QPushButton("Расширенный режим")
        self.BTN_extended_mode.setCheckable(True)

    def CB_create(self) -> None:
        self.CB_tp_run_infinit = QCheckBox("бесконечный цикл тестирования")
        self.CB_tc_run_single = QCheckBox("запускать только выбранный тесткейс")

        # SETTINGS -------------------------
        # self.CB_tp_run_infinit.setText("CB_text")
        # self.CB_tp_run_infinit.setText("CB_text")

    def PTE_create(self) -> None:
        self.PTE = QPlainTextEdit()

        # METHODS ORIGINAL ---------------------------------
        # self.PTE.setEnabled(True)
        # self.PTE.setUndoRedoEnabled(True)
        # self.PTE.setReadOnly(True)
        # self.PTE.setMaximumBlockCount(15)
        self.PTE.setMaximumWidth(500)

        # self.PTE.clear()
        self.PTE.appendPlainText(PTE_RESULTS_EXAMPLE)

        # self.PTE.appendHtml("")
        # self.PTE.anchorAt(#)
        # self.PTE.setSizeAdjustPolicy(#)
        # self.PTE.setHidden(True)

        # METHODS COMMON -----------------------------------
        self.PTE.setFont(QFont("Calibri (Body)", 7))

    def TV_create(self):
        # TODO: move examples to pyqtTemplate!
        self.TM = TpTableModel(self.DATA)

        self.TV = QTableView()
        self.TV.setModel(self.TM)
        self.TV.setSelectionMode(QTableView.SingleSelection)

        hh: QHeaderView = self.TV.horizontalHeader()
        # hh.setSectionHidden(self.TM.ADDITIONAL_COLUMNS - 1, True)
        hh.setSectionsClickable(False)
        # hh.setStretchLastSection(True)
        hh.setSectionResizeMode(QHeaderView.ResizeToContents)   # autoresize width!

        # hh.setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)  # for index of column set stretch size
        # hh.setMinimumSize(0, QHeaderView.ResizeMode.Stretch)
        hh.setMinimumWidth(1000)    # dont understand it is not work at least with ResizeToContents

        self.TV.resizeColumnsToContents()
        self.TV.resizeRowsToContents()


    # SLOTS ===========================================================================================================
    def slots_connect(self):
        self.CB_tp_run_infinit.stateChanged.connect(self.CB_tp_run_infinit__changed)
        self.CB_tc_run_single.stateChanged.connect(self.CB_tc_run_single__changed)

        self.BTN_start.toggled.connect(self.BTN_start__toggled)
        self.BTN_settings.toggled.connect(self.BTN_settings__toggled)
        self.BTN_devs_detect.clicked.connect(self.BTN_devs_detect__clicked)
        self.BTN_save.clicked.connect(self.BTN_save__clicked)
        self.BTN_clear_all.clicked.connect(self.BTN_clear_all__clicked)
        self.BTN_reset_all.clicked.connect(self.BTN_reset_all__clicked)
        self.BTN_extended_mode.toggled.connect(self.BTN_extended_mode__toggled)

        # self.DATA.signal__tp_finished.connect(self.BTN_reset_all__clicked)
        self.DATA.signal__tp_finished.connect(lambda: self.BTN_start.setChecked(False))
        self.DATA.signal__tp_finished.connect(self.TM._data_reread)

        self.DATA.signal__devs_detected.connect(self.DIALOGS.finished__devs_detection)
        self.DATA.signal__tp_finished.connect(self.DIALOGS.finished__tp)

        TestCaseBase.signals.signal__tc_state_changed.connect(lambda _: self.TM._data_reread())

        self.TV.selectionModel().selectionChanged.connect(self.TV_selectionChanged)
        self.TV.horizontalHeader().sectionClicked.connect(self.TV_hh_sectionClicked)

    # -----------------------------------------------------------------------------------------------------------------
    def CB_tp_run_infinit__changed(self, state: Optional[int] = None) -> None:
        """
        :param state:
            0 - unchecked
            1 - halfCHecked (only if isTristate)
            2 - checked (even if not isTristate)
        """
        self.DATA.TP_RUN_INFINIT = bool(state)

    def CB_tc_run_single__changed(self, state: Optional[int] = None) -> None:
        """
        :param state:
            0 - unchecked
            1 - halfCHecked (only if isTristate)
            2 - checked (even if not isTristate)
        """
        if state:
            self.DATA._TC_RUN_SINGLE = True
        else:
            self.DATA._TC_RUN_SINGLE = False
            if not self.DATA.isRunning():
                self.DATA.tc_active = None

        self.TM._data_reread()

    def BTN_start__toggled(self, state: Optional[bool] = None) -> None:
        # print(f"btn {state=}")
        if self.DATA.isRunning():
            state = False

        if state:
            self.DATA.start()
        elif not state:
            # if not self.DATA.isFinished():
            self.DATA.terminate()

        self.TM._data_reread()

    def BTN_settings__toggled(self, state: Optional[bool] = None) -> None:
        # print(f"BTN_select_tc_on_duts__toggled {state=}")
        # self.TV.horizontalHeader().setSectionHidden(self.TM.ADDITIONAL_COLUMNS - 1, not state)
        self.TV.horizontalHeader().setSectionsClickable(state)
        self.TM.open__settings = state
        self.TM._data_reread()

    def BTN_devs_detect__clicked(self) -> None:
        self.DATA.DEVICES__BREEDER_CLS.group_call__("address_forget")
        self.DATA.DEVICES__BREEDER_CLS.CLS_LIST__DUT.ADDRESSES__SYSTEM.clear()
        self.TM._data_reread()
        self.DATA.DEVICES__BREEDER_CLS.group_call__("address__resolve")    # MOVE TO THREAD??? no! not so need!
        self.TM._data_reread()
        self.DATA.signal__devs_detected.emit()

    def BTN_save__clicked(self) -> None:
        self.DATA.save__results()
        self.DIALOGS.finished__save()

    def BTN_reset_all__clicked(self) -> None:
        self.DATA.DEVICES__BREEDER_CLS.group_call__("reset")

    def BTN_clear_all__clicked(self) -> None:
        self.DATA.tcs_clear()
        self.TM._data_reread()

    def BTN_extended_mode__toggled(self, state: Optional[bool] = None) -> None:
        self.PTE.setHidden(not state)

        hh: QHeaderView = self.TV.horizontalHeader()
        hh.setSectionHidden(self.TM.HEADERS.SKIP, not state)
        hh.setSectionHidden(self.TM.HEADERS.ASYNC, not state)
        hh.setSectionHidden(self.TM.HEADERS.STARTUP_CLS, not state)
        hh.setSectionHidden(self.TM.HEADERS.TEARDOWN_CLS, not state)

    def TV_hh_sectionClicked(self, index: int) -> None:
        if index == self.TM.HEADERS.STARTUP_CLS:
            pass

        if index == self.TM.HEADERS.TEARDOWN_CLS:
            pass

        if index in self.TM.HEADERS.DUTS:
            dut = self.DATA.DEVICES__BREEDER_CLS.LIST__DUT[self.TM.HEADERS.DUTS.get_listed_index__by_outer(index)]
            dut.SKIP_reverse()
            self.TM._data_reread()

    def TV_selectionChanged(self, first: QItemSelection, last: QItemSelection) -> None:
        # print("selectionChanged")
        # print(f"{first=}")  # first=<PyQt5.QtCore.QItemSelection object at 0x000001C79A107460>
        # ObjectInfo(first.indexes()[0]).print(_log_iter=True, skip_fullnames=["takeFirst", "takeLast"])

        if not first:
            # when item with noFlag IsSelectable
            return

        index: QModelIndex = first.indexes()[0]

        row = index.row()
        col = index.column()

        try:
            tc_cls = list(self.DATA.TCS__CLS)[row]
        except:
            tc_cls = None

        row_is_summary: bool = tc_cls is None

        if self.DATA._TC_RUN_SINGLE and not row_is_summary:
            self.DATA.tc_active = tc_cls

        if col == self.TM.HEADERS.STARTUP_CLS:
            if not row_is_summary:
                self.PTE.setPlainText(str(tc_cls.result__startup_cls))

        if col in self.TM.HEADERS.DUTS:
            if row_is_summary:
                pass    # TODO: add summary_result
            else:
                dut = self.DATA.DEVICES__BREEDER_CLS.LIST__DUT[col - self.TM.HEADERS.DUTS.START_OUTER]
                self.PTE.setPlainText(tc_cls.TCS__LIST[dut.INDEX].get__results_pretty())

        if col == self.TM.HEADERS.TEARDOWN_CLS:
            if not row_is_summary:
                self.PTE.setPlainText(str(tc_cls.result__teardown_cls))

        self.TM._data_reread()


# =====================================================================================================================



from typing import *
import pathlib

from base_aux.base_statics.m1_types import *
from base_aux.base_inits.m1_nest_init_source import NestInit_Source
from base_aux.base_resolver.m1_resolver import NestCall_Resolve


# =====================================================================================================================
@final
class Resolve_DirPath(NestInit_Source, NestCall_Resolve):
    """
    GOAL
    ----
    resolve dirpath by draft

    SPECIALLY CREATED FOR
    ---------------------
    Resolve_FilePath init dirpath
    """
    SOURCE: TYPING.PATH_DRAFT | None

    def resolve(self) -> TYPING.PATH_FINAL:
        if self.SOURCE is not None:
            return pathlib.Path(self.SOURCE)
        if self.SOURCE is None:
            return pathlib.Path().cwd()


# =====================================================================================================================



import time
from typing import *
import pytest

from base_aux.threads.m1_threads import *


# =====================================================================================================================
class Test__ThreadItem:
    # -----------------------------------------------------------------------------------------------------------------
    def setup_method(self, method):
        # self.victim = ThreadItem()
        pass

    # -----------------------------------------------------------------------------------------------------------------
    def test_SecondaryStart(self):
        def target():
            time.sleep(0.2)

        victim = ThreadItem(target=target)
        victim.start()
        victim.wait()
        victim.start()
        victim.wait()
        assert True


# =====================================================================================================================
class Test__Manager:
    # VICTIM: type[ThreadsManager] = type("VICTIM", (ThreadsManager,), {})

    def test__singleton(self):
        class ThreadManager1(ThreadsManager):
            pass

        class ThreadManager2(ThreadsManager):
            pass

        inst1 = ThreadManager1()
        inst2 = ThreadManager2()
        assert id(inst1) != id(inst2)
        assert id(inst1) == id(ThreadManager1())

    # -----------------------------------------------------------------------------------------------------------------
    def test__noClass_severalManagers(self):
        # settings ------------------
        count = 5
        time_start = time.time()

        # define victim ------------------
        class ThreadManager1(ThreadsManager):
            pass

        class ThreadManager2(ThreadsManager):
            pass

        assert ThreadManager1() != ThreadManager2()
        assert ThreadManager1() is not ThreadManager2()

        assert ThreadManager1().count == 0
        assert ThreadManager2().count == 0

        @ThreadManager1().decorator__to_thread
        def func1(num):
            time.sleep(1)
            return num * 10

        @ThreadManager2().decorator__to_thread
        def func2(num):
            time.sleep(1)
            return num * 100

        # spawn ------------------
        for i in range(count):
            assert func1(i) is None

        assert ThreadManager1().count == count
        assert ThreadManager2().count == 0

        for i in range(count):
            assert func2(i) is None

        assert ThreadManager1().count == count
        assert ThreadManager2().count == count

        # wait ------------------
        ThreadManager1().wait_all()
        ThreadManager2().wait_all()

        # checks ------------------
        for item in ThreadManager1().THREADS:
            assert item.isRunning() is False

        assert time.time() - time_start < 5

        assert {item.result for item in ThreadManager1().THREADS} == {num * 10 for num in range(count)}
        assert {item.result for item in ThreadManager2().THREADS} == {num * 100 for num in range(count)}

    def test__Class(self):
        # settings ------------------
        count = 5
        time_start = time.time()

        # define victim ------------------
        class ThreadManager1(ThreadsManager):
            pass

        class Cls:
            @ThreadManager1().decorator__to_thread
            def func1(self, num):
                time.sleep(1)
                return num * 1000

        # spawn ------------------
        for i in range(count):
            assert Cls().func1(i) is None

        assert ThreadManager1().count == count
        ThreadManager1().wait_all()
        assert {item.result for item in ThreadManager1().THREADS} == {num * 1000 for num in range(count)}

        ThreadManager1().clear()

        # spawn ------------------
        for i in range(count):
            assert Cls().func1(i) is None

        assert ThreadManager1().count == count
        ThreadManager1().wait_all()
        assert {item.result for item in ThreadManager1().THREADS} == {num * 1000 for num in range(count)}

    def test__check_results_all(self):
        # define victim ------------------
        class ThreadManager1(ThreadsManager):
            pass

        @ThreadManager1().decorator__to_thread
        def func1(value):
            return value

        # bool ----------
        ThreadManager1().clear()
        [func1(True), func1(True)]
        ThreadManager1().wait_all()
        assert ThreadManager1().check_results_all() is True

        ThreadManager1().clear()
        [func1(True), func1(False)]
        ThreadManager1().wait_all()
        assert ThreadManager1().check_results_all() is False

        ThreadManager1().clear()
        [func1(False), func1(False)]
        ThreadManager1().wait_all()
        assert ThreadManager1().check_results_all(False) is True

        # int ----------
        ThreadManager1().clear()
        [func1(1), func1(1)]
        ThreadManager1().wait_all()
        assert ThreadManager1().check_results_all(1) is True

        ThreadManager1().clear()
        [func1(1), func1(2)]
        ThreadManager1().wait_all()
        assert ThreadManager1().check_results_all(1) is False

        # func_validate ----------
        ThreadManager1().clear()
        [func1(0), func1(1)]
        ThreadManager1().wait_all()
        assert ThreadManager1().check_results_all(func_validate=bool) is False

        ThreadManager1().clear()
        [func1(1), func1(2)]
        ThreadManager1().wait_all()
        assert ThreadManager1().check_results_all(func_validate=bool) is True

        def validate_int(obj: Any) -> bool:
            return isinstance(obj, int)

        ThreadManager1().clear()
        [func1(0), func1(1)]
        ThreadManager1().wait_all()
        assert ThreadManager1().check_results_all(func_validate=validate_int) is True

    def test__PARAM__NOTHREAD(self):
        # define victim ------------------
        class ThreadManager1(ThreadsManager):
            pass

        @ThreadManager1().decorator__to_thread
        def func1(value):
            time.sleep(0.2)
            return value

        # bool ----------
        ThreadManager1().clear()

        assert func1(True, nothread=False) is None
        assert func1(True, nothread=True) is True
        assert ThreadManager1().count == 2

    def test__AS_FUNC(self):
        class ThreadManager1(ThreadsManager):
            pass

        def func1(value):
            time.sleep(0.2)
            return value

        ThreadManager1().clear()
        thread = ThreadManager1().decorator__to_thread(func1)

        assert thread(True, nothread=False) is None
        assert thread(True, nothread=True) is True
        assert ThreadManager1().count == 2

    def _test__twice_execute_7777(self):    # not expected???
        class ThreadManager1(ThreadsManager):
            pass

        @ThreadManager1().decorator__to_thread
        def func1(value):
            time.sleep(0.2)
            return value

        ThreadManager1().clear()

        assert ThreadManager1().count == 0
        assert func1(True) is None
        assert ThreadManager1().count == 1
        ThreadManager1().wait_all()


# =====================================================================================================================



# =====================================================================================================================
"""
THIS IS THE REAL TESTPLAN!!!
"""
from typing import *
import time
import json
from pathlib import Path

from PyQt5.QtCore import QThread, pyqtSignal
from importlib import import_module

from base_aux.servers.m1_client_requests import *
from base_aux.aux_datetime.m1_datetime import *
from base_aux.loggers.m1_logger import *


# =====================================================================================================================
from .tc import TestCaseBase
from .devices import DutBase, DeviceBase, DevicesBreeder_WithDut, DevicesBreeder_Example
from .gui import TpGuiBase
from .api import TpApi_FastApi


# =====================================================================================================================
class Exx__TcsPathNotExists(Exception):
    pass


class Exx__TcItemNotFound(Exception):
    pass


class Exx__TcItemType(Exception):
    pass


class Exx__TcSettingsIncorrect(Exception):
    pass


# =====================================================================================================================
class TpMultyDutBase(Logger, QThread):
    signal__tp_start = pyqtSignal()
    signal__tp_stop = pyqtSignal()
    signal__tp_finished = pyqtSignal()
    signal__devs_detected = pyqtSignal()

    _signal__tp_reset_duts_sn = pyqtSignal()

    # SETTINGS ------------------------------------------------------
    TP_RUN_INFINIT: bool | None = None     # True - when run() started - dont stop!
    TP_RUN_INFINIT__TIMEOUT: int = 1

    _TC_RUN_SINGLE: bool | None = None

    START__GUI_AND_API: bool = True

    STAND_NAME: Optional[str] = "stand_id__1"
    STAND_DESCRIPTION: Optional[str] = "stand_description"
    STAND_SN: Optional[str] = "StandSn"

    API_SERVER__START: bool = True
    API_SERVER__CLS: type[TpApi_FastApi] = TpApi_FastApi
    api_server: TpApi_FastApi

    GUI__START: bool = True
    GUI__CLS: type[TpGuiBase] = TpGuiBase

    api_client: Client_RequestsStack = Client_RequestsStack()   # todo: USE CLS!!! + add start

    # DIRPATH_TPS: Union[str, Path] = "TESTPLANS"
    DIRPATH_TCS: Union[str, Path] = "TESTCASES"
    DIRPATH_RESULTS: Union[str, Path] = "RESULTS"
    # DIRPATH_DEVS: Union[str, Path] = "DEVICES__BREEDER_INST"
    SETTINGS_BASE_NAME: Union[str, Path] = "SETTINGS_BASE.json"
    SETTINGS_BASE_FILEPATH: Path

    DEVICES__BREEDER_CLS: type[DevicesBreeder_WithDut] = DevicesBreeder_Example

    # AUX -----------------------------------------------------------
    TCS__CLS: dict[Union[str, type[TestCaseBase]], Optional[bool]] = {}     # todo: RENAME TO clss!!!
    # {
    #     Tc1: True,
    #     Tc2: True
    # }

    # DEVICES__BREEDER_INST: list[Union[str, type[DeviceBase]]]    # settings
    # [
    #     Dev1,
    #     Dev2
    # ]

    __tc_active: Optional[type[TestCaseBase]] = None
    tc_prev: Optional[type[TestCaseBase]] = None
    progress: int = 0   # todo: use as property? by getting from TCS???

    # =================================================================================================================
    @property
    def tc_active(self) -> type[TestCaseBase] | None:
        return self.__tc_active

    @tc_active.setter
    def tc_active(self, value: type[TestCaseBase] | None) -> None:
        if self.__tc_active and not self.__tc_active.SKIP:
            self.tc_prev = self.__tc_active
        self.__tc_active = value

    def tp__check_active(self) -> bool:
        result = self.tc_active is not None and self.progress not in [0, 100]
        return result

    # =================================================================================================================
    def __init__(self):
        super().__init__()
        # self.DIRPATH_TPS: Path = Path(self.DIRPATH_TPS)
        self.DIRPATH_TCS: Path = Path(self.DIRPATH_TCS)
        # self.DIRPATH_DEVS: Path = Path(self.DIRPATH_DEVS)
        self.SETTINGS_BASE_FILEPATH = self.DIRPATH_TCS.joinpath(self.SETTINGS_BASE_NAME)

        self.DIRPATH_RESULTS = pathlib.Path(self.DIRPATH_RESULTS)
        if not self.DIRPATH_RESULTS.exists():
            self.DIRPATH_RESULTS.mkdir(parents=True, exist_ok=True)

        if not self.DIRPATH_TCS.exists():
            msg = f"[ERROR] not found path {self.DIRPATH_TCS.name=}"
            print(msg)
            raise Exx__TcsPathNotExists(msg)

        self.DEVICES__BREEDER_CLS.generate__objects()

        self.tcs__reinit()
        self.slots_connect()

        if self.START__GUI_AND_API:
            self.start__gui_and_api()

    def start__gui_and_api(self) -> None:
        if self.API_SERVER__START:
            self.LOGGER.debug("starting api server")
            self.api_server = self.API_SERVER__CLS(data=self)
            self.api_server.start()

        # last execution --------------------------------------
        if self.GUI__START:
            self.LOGGER.debug("starting gui")
            self.gui = self.GUI__CLS(self)

            # this will BLOCK process
            # this will BLOCK process
            # this will BLOCK process
            # this will BLOCK process
            # this will BLOCK process
            self.gui.run()
        elif self.API_SERVER__START:
            self.api_server.wait()  # it is ok!!!

    def slots_connect(self) -> None:
        self.signal__tp_start.connect(self.start)
        self.signal__tp_stop.connect(self.terminate)
        self._signal__tp_reset_duts_sn.connect(self.DEVICES__BREEDER_CLS._debug__duts__reset_sn)

        TestCaseBase.signals.signal__tc_state_changed.connect(self.post__tc_results)

    # =================================================================================================================
    def tcs__reinit(self) -> None:
        if not self.TCS__CLS:
            self._tcs__load()
        self._tcs__apply_classes()
        self._tcs__apply_settings()
        self._tcs__apply_devices()

    def _tcs__load(self) -> None:
        """
        for tests just overwrite
        :return:
        """
        self._tcs__load_from_files()

    def _tcs__load_from_files(self) -> None:
        self.TCS__CLS = {}
        for file in self.DIRPATH_TCS.glob("*.py"):
            if not file.stem.startswith("__"):
                self.TCS__CLS.update({file.stem: True})

    def _tcs__apply_classes(self) -> None:
        result = {}
        for item, using in self.TCS__CLS.items():
            # print(dir(TESTCASES))
            print(f"touch {self.DIRPATH_TCS} {item=}")
            if isinstance(item, str):   # filename
                # tc_cls = import_module(item, "TESTCASES").TestCase    # not working!
                # tc_cls = getattr(TESTCASES, item).TestCase      # not working
                tc_cls = None
                try:
                    tc_cls = import_module(f"{self.DIRPATH_TCS.name}.{item}").TestCase
                except Exception as exx:
                    msg = f"[WARN] no 'TestCase' class in file [{self.DIRPATH_TCS}/{item}] {exx=}"
                    print(msg)
                    continue
                if not tc_cls:
                    msg = f"[ERROR] file not found[{item=}] in /{self.DIRPATH_TCS}/"
                    raise Exx__TcItemNotFound(msg)
                tc_cls.NAME = item
            elif isinstance(type(item), type) and issubclass(item, TestCaseBase):
                tc_cls = item
                # msg = f"[ERROR] DONT USE IT!"
                # raise Exception(msg)
            else:
                msg = f"[ERROR] type is inconvenient [{item=}]"
                raise Exx__TcItemType(msg)

            tc_cls.SKIP = not using
            result.update({tc_cls: using})

        self.TCS__CLS = result

    def _tcs__apply_settings(self) -> None:
        for tc_cls in self.TCS__CLS:
            tc_cls.SETTINGS_FILES = [self.SETTINGS_BASE_FILEPATH, ]

            settings_tc_filepath = self.DIRPATH_TCS.joinpath(f"{tc_cls.NAME}.json")
            if settings_tc_filepath.exists():
                tc_cls.SETTINGS_FILES.append(settings_tc_filepath)
            else:
                print(f"{settings_tc_filepath=} NOT_EXISTS")
                pass

        # print(f"{tc_cls.SETTINGS=}")

    def _tcs__apply_devices(self) -> None:
        for tc in self.TCS__CLS:
            tc.devices__apply(self.DEVICES__BREEDER_CLS)

    def tcs_clear(self) -> None:
        for tc_cls in self.TCS__CLS:
            tc_cls.clear__cls()

    # =================================================================================================================
    def tp__startup(self) -> bool:
        """
        Overwrite with super! super first!
        """
        self.tc_prev = None
        self.progress = 1
        self.DEVICES__BREEDER_CLS.group_call__("connect__only_if_address_resolved")  #, group="DUT")   # dont connect all here! only in exact TC!!!!????
        return True

    def tp__teardown(self, progress: int = 100) -> None:
        """
        Overwrite with super! super last!
        """
        if self.tc_active and not self.tc_active.finished:
            self.tc_active.terminate__cls()
        if self.tc_prev and not self.tc_prev.finished:
            self.tc_prev.teardown__cls()
            self.tc_prev = None
        if not self._TC_RUN_SINGLE:
            self.tc_active = None

        if progress is None:
            progress = 100
        self.progress = progress

        self.DEVICES__BREEDER_CLS.disconnect__cls()

        # self.signal__tp_finished.emit()   # dont place here!!!

    # =================================================================================================================
    def terminate(self) -> None:
        pass

        need_msg: bool = False
        if self.isRunning():
            need_msg = True
            super().terminate()

        # TERMINATE CHILDS!!! ---------------------
        # ObjectInfo(self.currentThread()).print()    # cant find childs!!!

        # finish active ----------------------------
        if self.tc_active:
            self.tc_active.terminate__cls()

        # finish ----------------------------
        self.tp__teardown(0)
        if need_msg:
            self.signal__tp_finished.emit()

    def run(self) -> None:
        self.LOGGER.debug("TP START")
        if self.tp__check_active():
            return

        cycle_count = 0
        while True:
            if not self._TC_RUN_SINGLE:
                self.tcs_clear()

            cycle_count += 1

            if self.tp__startup():
                if self._TC_RUN_SINGLE:
                    if self.tc_active:
                        self.tc_active.run__cls(single=True)
                else:
                    for self.tc_active in filter(lambda x: not x.SKIP, self.TCS__CLS):  # TODO: place cls_prev into TcBaseCls!!! and clear on finish???
                        tc_executed__result = self.tc_active.run__cls(cls_prev=self.tc_prev)
                        if tc_executed__result is False:
                            break
            # if self.tc_prev:
            #     self.tc_prev.teardown__cls()
            if self.tc_active and self.tc_active.finished__cls is not None:
                self.tc_active.teardown__cls()
            # FINISH TP CYCLE ---------------------------------------------------
            self.tp__teardown()
            self.LOGGER.debug("TP FINISH")

            # RESTART -----------------------------------------------------
            if not self.TP_RUN_INFINIT:
                break

            time.sleep(self.TP_RUN_INFINIT__TIMEOUT)

        # FINISH TP TOTAL ---------------------------------------------------
        self.signal__tp_finished.emit()

    # =================================================================================================================
    def get__info__stand(self) -> dict[str, Any]:
        result = {
            "STAND_NAME": self.STAND_NAME,
            "STAND_DESCRIPTION": self.STAND_DESCRIPTION,
            "STAND_SN": self.STAND_SN,
            "STAND_SETTINGS": TestCaseBase.settings_read(files=self.SETTINGS_BASE_FILEPATH),
        }
        return result

    def get__info__tp(self) -> dict[str, Any]:
        """
        get info/structure about stand/TP
        """
        TP_TCS = []
        for tc in self.TCS__CLS:
            TP_TCS.append(tc.get__info__tc())

        result = {
            **self.get__info__stand(),

            "TESTCASES": TP_TCS,
            # "TP_DUTS": [],      # TODO: decide how to use
            # [
            #     # [{DUT1}, {DUT2}, …]
            #     {
            #         DUT_ID: 1  # ??? 	# aux
            #         DUT_SKIP: False
            #     }
            # ]

            }
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def get__results(self) -> dict[str, Any]:
        """
        get all results for stand/TP
        """
        TCS_RESULTS = {}
        for tc_cls in self.TCS__CLS:
            TCS_RESULTS.update({tc_cls: tc_cls.get__results__all()})

        result = {
            "STAND" : self.get__info__stand(),
            "TCS": TCS_RESULTS,
        }
        return result

    def save__results(self) -> None:
        name_prefix = str(DateTimeAux())
        for index in range(self.DEVICES__BREEDER_CLS.COUNT):
            result_i = {}
            for tc_cls in self.TCS__CLS:
                tc_inst = None
                try:
                    tc_inst: TestCaseBase = tc_cls.TCS__LIST[index]
                    tc_inst_result = tc_inst.get__results(add_info_dut=False, add_info_tc=False)["tc_result"]
                except:
                    tc_inst_result = None

                result_i.update({tc_cls.DESCRIPTION: tc_inst_result})

            dut_info = tc_inst.DEVICES__BREEDER_INST.DUT.get__info__dev()
            result_dut = {
                "STAND": self.get__info__stand(),
                "DUT": dut_info,
                "RESULTS": result_i,
            }
            data_text = json.dumps(result_dut, indent=4, ensure_ascii=False)

            filename = f"{name_prefix}[{index}].json"
            filepath = pathlib.Path(self.DIRPATH_RESULTS, filename)
            filepath.write_text(data=data_text, encoding='utf-8')

    # -----------------------------------------------------------------------------------------------------------------
    def post__tc_results(self, tc_inst: TestCaseBase) -> None:
        # CHECK ------------------------------------------
        if not self.api_client or tc_inst.result is None:
            return

        # WORK ------------------------------------------
        try:
            tc_results = tc_inst.get__results()
        except:
            tc_results = {}

        body = {
            **self.get__info__stand(),
            **tc_results,
        }
        self.api_client.send(body=body)


# =====================================================================================================================
class TpInsideApi_Runner(TpApi_FastApi):
    """
    REASON:
    in windows TestCaseBase works fine by any variance GUI__START/API_SERVER__START
    in Linux it is not good maybe cause of nesting theme=Thread+Async+Threads

    so this is the attempt to execute correctly TP in Linux by deactivating GUI and using theme=Async+Threads

    UNFORTUNATELY: ITS NOT WORKING WAY for linux!!!
    """
    TP_CLS: type[TpMultyDutBase] = TpMultyDutBase

    def __init__(self, *args, **kwargs):

        self.TP_CLS.START__GUI_AND_API = False
        self.data = self.TP_CLS()

        super().__init__(*args, **kwargs)
        self.run()


# =====================================================================================================================



from base_aux.base_statics.m1_types import TYPES
from base_aux.base_inits.m1_nest_init_source import *


# =====================================================================================================================
@final
class TypeAux(NestInit_Source):
    SOURCE: type[Any] | Any

    # -----------------------------------------------------------------------------------------------------------------
    def check__bool_none(self) -> bool:
        """
        GOAL
        ----
        help in case of
            assert 0 == False
            assert 1 == True
            assert 2 == False   # unclear!!!

        CREATED SPECIALLY FOR
        ---------------------
        funcs.Valid.compare_doublesided
        """
        return isinstance(self.SOURCE, (bool, type(None)))

    def check__elementary(self) -> bool:
        if callable(self.SOURCE):
            return False
        return isinstance(self.SOURCE, TYPES.ELEMENTARY)

    def check__elementary_single(self) -> bool:
        return isinstance(self.SOURCE, TYPES.ELEMENTARY_SINGLE)

    def check__elementary_single_not_none(self) -> bool:
        """
        its just an Idea!!!

        GOAL
        ----
        prepare to work with ensure_collection
        None assumed as not Passed value! so we can ensure to return None -> ()

        SPECIALLY CREATED FOR
        ---------------------
        ensure_collection somewhere!
        """
        return self.check__elementary_single() and self.SOURCE is not None

    def check__elementary_collection(self) -> bool:
        """
        GOAL
        ----
        MOST PREFERRED to use for ensure_collection! and apply for Args!
        all other aux_types (ClsInst) will covered by brackets!
        """
        return isinstance(self.SOURCE, TYPES.ELEMENTARY_COLLECTION)

    def check__elementary_collection_not_dict(self) -> bool:
        return isinstance(self.SOURCE, TYPES.ELEMENTARY_COLLECTION) and not isinstance(self.SOURCE, dict)

    # -----------------------------------------------------------------------------------------------------------------
    def check__module(self) -> bool:
        """
        GOAL
        ----
        check source is module

        EXPLORE
        -------
            import sys

            print(type(sys))  # <class 'module'>
            print(type(int))  # <class 'type'>

            # print(isinstance(sys, module))       # NameError: name 'module' is not defined

            # import module       # ModuleNotFoundError: No module named 'module'
            # print(isinstance(sys, module))       # NameError: name 'module' is not defined

            def check_obj_is_module(source) -> bool:
                result = str(type(source)) == "<class 'module'>"
                print(f"{source=}/{result=}")
                return result

            for obj in [int, sys, 1]:
                check_obj_is_module(obj)

            source=<class 'int'>/result=False
            source=<module 'sys' (built-in)>/result=True
            source=1/result=False
        """
        result = str(type(self.SOURCE)) == "<class 'module'>"
        # print(f"{self.SOURCE=}/{result=}")
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def check__iterable(
            self,
            dict_as_iterable: bool = True,
            str_and_bytes_as_iterable: bool = True,
    ) -> bool:
        """checks if SOURCE is iterable.

        :param dict_as_iterable: if you dont want to use dict in your selecting,
            becouse maybe you need flatten all elements in list/set/tuple into one sequence
            and dict (as extended list) will be irrelevant!
        :param str_and_bytes_as_iterable: usually in data processing you need to work with str-type elements as OneSolid element
            but not iterating through chars!
        """
        if isinstance(self.SOURCE, dict):
            return dict_as_iterable
        elif isinstance(self.SOURCE, (str, bytes)):
            return str_and_bytes_as_iterable
        elif isinstance(self.SOURCE, (tuple, list, set,)):  # need to get it explicitly!!!
            return True
        elif hasattr(self.SOURCE, '__iter__') or hasattr(self.SOURCE, '__getitem__'):
            return True

        # FINAL ---------------------
        return False

    def check__iterable_not_str(self) -> bool:
        """
        GOAL
        ----
        checks if SOURCE is iterable, but not exactly str!!!
        """
        return self.check__iterable(str_and_bytes_as_iterable=False)

    # CALLABLE --------------------------------------------------------------------------------------------------------
    def check__callable_func_meth_inst_cls(self) -> bool:
        """
        GOAL
        ----
        just any callable or CLASS!!! - so it is actually all CALLABLE!


        CREATES SPECIALLY FOR
        ---------------------
        just to see difference and clearly using by name!!!
        """
        return callable(self.SOURCE)

    def check__callable_func_meth_inst(self) -> bool:
        """
        GOAL
        ----
        just any callable but NO CLASS!!!


        CREATES SPECIALLY FOR
        ---------------------
        detect all funcs like func/meth/or even DescriptedClasses (it is class but actually used like func!)
        recommended using instead of just Callable! cause Callable keeps additionally every class instead of just simple func/method!
        """
        if self.check__class():
            result = issubclass(self.SOURCE, TYPES.ELEMENTARY)
        else:
            result = callable(self.SOURCE)
        return result

    def check__callable_func_meth(self) -> bool:
        return self.check__callable_func() or self.check__callable_meth()

    def check__callable_func(self) -> bool:
        """
        if only the exact generic function! no class method!
        Lambda included!
        Classmethod included!

        CREATED SPECIALLY FOR
        ---------------------
        not special! just as ones found ability to!
        """
        if self.check__callable_cls_as_func_builtin():
            result = True
        else:
            result = TYPES.FUNCTION in self.SOURCE.__class__.__mro__
        return result

    def check__callable_meth(self) -> bool:
        """
        if only the exact instance method!
        no generic funcs!
        no CALLABLE INSTANCE!
        no callable classes!

        CREATED SPECIALLY FOR
        ---------------------
        not special! just as ones found ability to!
        """
        result = not self.check__class() and TYPES.METHOD in self.SOURCE.__class__.__mro__
        return result

    def check__callable_inst(self) -> bool:
        """
        CREATED SPECIALLY FOR
        ---------------------
        not special! just as ones found ability to!
        """
        result = self.check__instance() and hasattr(self.SOURCE, "__call__")
        return result

    def check__callable_cls_as_func_builtin(self) -> bool:
        """
        if class and class is as func like int/str/*  or nested
        """
        return self.check__class() and issubclass(self.SOURCE, TYPES.ELEMENTARY)

    # CLS/INST --------------------------------------------------------------------------------------------------------
    def check__class(self) -> bool:
        """
        works both for funcs/meths for any Сды/Штые1 see tests test__check__class
        """
        # 1 -----------------
        # return hasattr(self.SOURCE, "__class__")     # this is incorrect!!! tests get fail!

        # 2 -----------------
        # INCORRECT! metaclasses have middle class(MetaType)!
        # if self.SOURCE.__class__ is type:
        #     return True

        # 3 -----------------
        # CORRECT! only Class can be subclass!!! otherwise Raise!
        try:
            return issubclass(self.SOURCE, object)
        except:
            return False

        # 4 -----------------
        # CORRECT! only Class have __mro__!!! otherwise Raise!
        # try:
        #     _ = self.SOURCE.__mro__
        #     return True
        # except:
        #     return False

    def check_metaclassed(self) -> bool:
        """
        just an idea to show that class crates by metaclass!
        no idea how and where to use! just a toy for playing and deeper understanding
        """
        # todo: finish

    def check__instance(self) -> bool:
        return not self.check__class() and not self.check__callable_func() and not self.check__callable_meth()

    def check__instance_not_elementary(self) -> bool:
        return self.check__instance() and not self.check__elementary()

    def check__nested__by_cls_or_inst(self, parent: Any) -> bool | None:
        """
        any of both variant (Instance/Class) comparing with TARGET of both variant (Instance/Class)

        specially created for pytest_aux for comparing with Exception!
        """
        source_cls = self.get__class()
        parent_cls = TypeAux(parent).get__class()
        return issubclass(source_cls, parent_cls)

    # EXX -------------------------------------------------------------------------------------------------------------
    def check__exception(self) -> bool:
        """
        any of both variant (Instance/Class) of any Exception!
        """
        try:
            return issubclass(self.SOURCE, Exception)   # check it first!!! in try cover
        except:
            return isinstance(self.SOURCE, Exception)

        # if isinstance(self.SOURCE, Exception):
        #     return True
        # try:
        #     return issubclass(self.SOURCE, Exception)
        # except:
        #     pass
        # return False

    # =================================================================================================================
    def get__class(self) -> type:
        """
        GOAL
        ----
        get class from any object

        CREATED SPECIALLY FOR
        ---------------------
        classes.ClsMiddleGroup
        """
        if self.check__class():
            return self.SOURCE
        else:
            return self.SOURCE.__class__

    def get_mro(self) -> tuple[type, ...]:
        """
        GAOL
        ----
        get DIRECT mro for instance/class!
        """
        cls = self.get__class()
        mro = cls.__mro__
        return mro

    def iter_mro_user(self, *exclude: type) -> Iterable[type]:
        """
        GAOL
        ----
        iter only user classes
        """
        mro = self.get_mro()
        if not mro:
            """
            created specially for
            ---------------------
            DictAttrAnnotRequired(dict)
            it is not working without it!!!
            """
            mro = ()

        for cls in mro:
            if cls not in [*exclude, object, *TYPES.ELEMENTARY]:
                yield cls

    # =================================================================================================================
    def type__init_value__default(self) -> Any | NoReturn:
        """
        GOAL
        ----
        gen default values by type if available

        SPECIALLY CREATED FOR
        ---------------------
        AnnotAux.init_values
        """
        source: type[Any] = self.SOURCE  # self.get__class() # DONT USE!

        if source in [type(None), None]:
            return None

        if source in [
            bool,
            int, float,
            str, bytes,

            list,
            tuple,  # TODO: resolve what to do with tuples!
            set, dict,
        ]:
            return source()

        if source.__module__ == "typing":
            if str(source).startswith("typing.Optional"):
                return None
            if str(source).startswith("typing.Union"):
                return TypeAux(source.__args__[0]).type__init_value__default()
            if (
                    str(source).startswith("typing.Iterable")
                    or
                    str(source).startswith("typing.Generator")
                    or
                    str(source).startswith("typing.Container")
            ):
                return []

            if (
                    # fixme: add Callable/other //... ||
                    str(source).startswith("typing.Callable")
                    or
                    str(source).startswith("typing.Any")
                    or
                    False
            ):
                raise NotImplemented()

        # FINAL ------------------------------------
        # USERCLASS
        if callable(source):
            return source
        else:
            return source


# =====================================================================================================================



from typing import *
from base_aux.base_inits.m1_nest_init_source import *
from base_aux.base_statics.m4_enums import *


# =====================================================================================================================
@final
class CallableAux(NestInit_Source):
    """
    """
    PROCESS: ProcessState = ProcessState.NONE

    def __call__(self, *args, **kwargs) -> Any | NoReturn:
        return self._construct_with_args_kwargs(*args, **kwargs)

    def _construct_with_args_kwargs(self, *args, **kwargs) -> Any | NoReturn:
        """
        unsafe (raise acceptable) get value
        """
        self.PROCESS: ProcessState = ProcessState.STARTED

        if callable(self.SOURCE):
            result = self.SOURCE(*args, **kwargs)
        else:
            result = self.SOURCE

        self.PROCESS: ProcessState = ProcessState.SUCCESS
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def resolve(self, callable_use: CallableResolve = CallableResolve.RAISE, *args, **kwargs) -> Any | None | Exception | NoReturn | CallableResolve | bool:
        """
        NOTE
        ----
        it is just a collection for all variants in one func!

        it is not so convenient to use param callable_use!
        SO preferred using other/further direct methods!
        """
        if callable_use == CallableResolve.DIRECT:
            return self.SOURCE

        elif callable_use == CallableResolve.EXX:
            return self.resolve_exx(*args, **kwargs)

        elif callable_use == CallableResolve.RAISE:
            return self.resolve_raise(*args, **kwargs)

        elif callable_use == CallableResolve.RAISE_AS_NONE:
            return self.resolve_raise_as_none(*args, **kwargs)

        elif callable_use == CallableResolve.SKIP_CALLABLE:
            return self.resolve_skip_callables(*args, **kwargs)

        elif callable_use == CallableResolve.SKIP_RAISED:
            return self.resolve_skip_raised(*args, **kwargs)

        elif callable_use == CallableResolve.BOOL:
            return self.resolve_bool(*args, **kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    def resolve_exx(self, *args, **kwargs) -> Any | Exception:
        """
        GOAL
        ----
        same as resolve_raise but
        attempt to simplify result by not using try-sentence.
        so if get raise in resolve_raise - return ClsException object

        USEFUL IDEA
        -----------
        1. in gui when its enough to get str() on result and see the result

        SPECIALLY CREATED FOR
        ---------------------
        just in case

        """
        try:
            return self(*args, **kwargs)
        except Exception as exx:
            self.PROCESS: ProcessState = ProcessState.RAISED
            return exx

    def resolve_raise(self, *args, **kwargs) -> Any | NoReturn:
        """
        just a direct result for call

        SPECIFIC LOGIC
        --------------
        if callable - call and return result.
        else - return source.

        GOAL
        ----
        get common expected for any python code result - simple calculate or raise!
        because of resolve_exx is not enough!

        CREATED SPECIALLY FOR
        ---------------------
        GetattrPrefixInst
        """
        return self(*args, **kwargs)

    def resolve_raise_as_none(self, *args, **kwargs) -> Any | None:
        try:
            return self.resolve_raise(*args, **kwargs)
        except:
            return None

    def resolve_skip_callables(self, *args, **kwargs) -> Any | NoReturn:
        if callable(self.SOURCE):
            return ProcessState.SKIPPED  # TODO: decide using None ???
        else:
            self.PROCESS: ProcessState = ProcessState.SKIPPED
            return self.SOURCE

    def resolve_skip_raised(self, *args, **kwargs) -> Any | NoReturn:
        try:
            return self.resolve_raise(*args, **kwargs)
        except:
            self.PROCESS: ProcessState = ProcessState.RAISED
            return ProcessState.SKIPPED  # TODO: decide using None ???

    def resolve_bool(self, *args, **kwargs) -> bool:
        """
        GOAL
        ----
        same as resolve_exx but
        apply bool func on result

        ability to get bool result with meanings:
            - methods/funcs must be called
                assert get_bool(LAMBDA_TRUE) is True
                assert get_bool(LAMBDA_NONE) is False

            - Exceptions assumed as False
                assert get_bool(Exception) is False
                assert get_bool(Exception("FAIL")) is False
                assert get_bool(LAMBDA_EXX) is False

            - for other values get classic bool()
                assert get_bool(None) is False
                assert get_bool([]) is False
                assert get_bool([None, ]) is True

                assert get_bool(LAMBDA_LIST) is False
                assert get_bool(LAMBDA_LIST, [1, ]) is True

            - if on bool() exception raised - return False!
                assert get_bool(ClsBoolRaise()) is False

        CREATED SPECIALLY FOR
        ---------------------
        funcs.Valid.skip_link or else value/func assumed as bool result
        """
        try:
            result = self.resolve_raise(*args, **kwargs)
            try:
                is_exx = issubclass(result, Exception)
            except:
                is_exx = isinstance(result, Exception)

            if is_exx:
                return False
            return bool(result)
        except:
            self.PROCESS: ProcessState = ProcessState.RAISED
            return False

    # -----------------------------------------------------------------------------------------------------------------
    def check_raise(self, *args, **kwargs) -> bool:
        """
        SPECIALLY CREATED FOR
        ---------------------
        check Privates in pytest for skipping

        USE LambdaTrySuccess instead!
        """
        try:
            self.resolve_raise(*args, **kwargs)
            return False
        except:
            self.PROCESS: ProcessState = ProcessState.RAISED
            return True

    def check_no_raise(self, *args, **kwargs) -> bool:
        return not self.check_raise(*args, **kwargs)


# =====================================================================================================================



from typing import *

import re
import sys
import glob
import time
import string
from enum import Enum, auto

from base_aux.loggers.m1_logger import Logger

from serial import Serial
from serial.tools import list_ports
from base_aux.aux_values.m3_unit import *
from base_aux.aux_cmp_eq.m2_eq_aux import EqAux

from .m0_history import HistoryIO


# =====================================================================================================================
class Exx_SerialAddress_NotApplyed(Exception):
    """
        raise SerialException("Port must be configured before it can be used.")
    serial.serialutil.SerialException: Port must be configured before it can be used.
    """
    pass


class Exx_SerialAddress_NotExists(Exception):
    """
    SerialException("could not open port 'COM6': FileNotFoundError(2, 'Не удается найти указанный файл.', None, 2)") - всегда несуществующий порт в Windows!!!
    """
    pass


class Exx_SerialAddresses_NoVacant(Exception):
    pass


class Exx_SerialAddresses_NoAutodetected(Exception):
    pass


class Exx_SerialAddress_AlreadyOpened(Exception):
    """
    SerialException("could not open port 'COM7': PermissionError(13, 'Отказано в доступе.', None, 5)")
    """
    pass


class Exx_SerialAddress_AlreadyOpened_InOtherObject(Exception):
    """
    SerialException("could not open port 'COM7': PermissionError(13, 'Отказано в доступе.', None, 5)")
    """
    pass


class Exx_SerialAddress_OtherError(Exception):
    """
    SerialException("could not open port 'COM7': OSError(22, 'Указано несуществующее устройство.', None, 433)") - порт есть, но получена ошибка при открытии!!!
    """
    pass


class Exx_SerialRead_NotFullLine(Exception):
    pass


class Exx_SerialRead_FailPattern(Exception):
    """
    if read string which match error pattern
    """
    pass


class Exx_SerialRead_FailDecoding(Exception):
    """
    REASON
    ------
    some serial devices (depends on microchips model) not always give correct reading bytes

    SOLVATION
    ---------
    1. [BESTandONLY] just use other device on other appropriate microchip!
    """


class Exx_SerialPL2303IncorrectDriver(Exception):
    """
    REASON
    ------
    typical for windows

    SOLVATION
    ---------
    1. [BEST] just use other device on other microchip
    2. manually select other more OLD driver in driver/device manager
        version 3.3.2.105/27.10.2008 - is good!
        version 3.8.22.0/22.02.2023 - is not good!!!
    """
    MARKER: str = "PL2303HXA PHASED OUT SINCE 2012. PLEASE CONTACT YOUR SUPPLIER"


# =====================================================================================================================
TYPE__RW_ANSWER_SINGLE = Union[None, str, ValueUnit]
TYPE__RW_ANSWER = Union[TYPE__RW_ANSWER_SINGLE, list[TYPE__RW_ANSWER_SINGLE]]


class Type__WrReturn(Enum):
    ALL_OUTPUT = auto()
    HISTORY_IO = auto()
    DICT = auto()


class Type__AddressAutoAcceptVariant(Enum):
    FIRST_FREE = auto()
    FIRST_FREE__SHORTED = auto()
    FIRST_FREE__PAIRED = auto()
    FIRST_FREE__ANSWER_VALID = auto()


TYPE__ADDRESS = Union[None, Type__AddressAutoAcceptVariant, str]


# =====================================================================================================================
class SerialClient(Logger):
    """
    GOOD ADAPTERS
    =============
    1. use good COM-port adapters!!!
        some bites may be lost (usually on started byte) or added extra chars (usually to start and end of line)!!!

        DECODING APPROPRIATE MODELS
        ---------------------------
        =WRONG=
        - driver PROFILIC PL2303(pcb SBT5329) - often +wrong driver
        - driver FTDI FT232RL(pcb msc-73lv) - sometimes but less than on Profilic=fail on step 0/11/20/3 3/9/95!!!
        - driver CP2102(pcb MostDense) - fail on step 3392/877/1141/!!! sometimes on step 1715 get SerialTimeoutException

        =GOOD=
        - driver CH340(pcb UsbToTtl) - no one error so far!
        - driver CH341A(pcb AllInOne/big universal) - no one error so far! more then steps about 50 minutes!!! tired of waiting
        - driver CH341A(pcb CH341T_V3) - no one error so far! more then steps about 35 minutes!!! tired of waiting

    CAREFUL
    -------
    IF YOU WANT TO BE SURE with WR methods with not good adapters
    use write_read__last_validate!!! it would rewrite data if not valid answer (even with incorrect but good decoding)!!!
    """
    pass
    pass
    pass

    # LOGGER --------------------------------------------------
    LOG_ENABLE = True

    # SETTINGS ------------------------------------------------
    _ADDRESS: TYPE__ADDRESS = None

    TIMEOUT__READ: float = 0.3  # 0.2 is too short!!! dont touch! in case of reading char by char 0.5 is the best!!! 0.3 is not enough!!!
    # need NONE NOT 0!!! if wait always!!
    BAUDRATE: int = 9600  # 115200

    REWRITEIF_READNOANSWER: int = 5
    REWRITEIF_READFAILDECODE: int = 5
    REWRITEIF_NOVALID: int = 5

    CMDS_DUMP: list[str] = []  # ["IDN", "ADR", "REV", "VIN", ]
    RAISE_CONNECT: bool = True
    RAISE_READ_FAIL_PATTERN: bool = False

    PREFIX: Optional[str] = None

    RELEASE_ON_DISCONNECT: bool | None = None

    # TODO: come up and apply ANSWER_SUCCESS??? may be not need cause of redundant
    ANSWER_SUCCESS: str = "OK"  # case insensitive
    ANSWER_FAIL_PATTERN: Union[str, list[str]] = [r".*FAIL.*", ]  # case insensitive!

    # rare INFRASTRUCTURE -----------------
    ENCODING: str = "utf8"
    EOL__SEND: bytes = b"\r\n"  # "\r"=ENTER in PUTTY  but "\r\n"=is better in read Putty!
    EOL__UNI_SET: bytes = b"\r\n"

    _GETATTR_STARTSWITH__SEND: str = "send__"
    _GETATTR_SPLITTER__ARGS: str = "__"

    # test purpose EMULATOR -----------------
    _EMULATOR__CLS: type[
        'SerialServer_Base'] = None  # IF USED - START it on PAIRED - it is exactly Emulator/Server! no need to use just another serialClient! _EMULATOR__INST could be used for test reason and check values in realtime!!
    _EMULATOR__INST: 'SerialServer_Base' = None
    _EMULATOR__START: bool = None  # DONT DELETE! it need when you reconnecting! cause of address replaced after disconnecting by exact str after PAIRED*

    # AUX -----------------------------------------------------
    history: HistoryIO = None
    _SERIAL: Serial

    ADDRESSES__SYSTEM: dict[str, Union[None, Self]] = {}
    ADDRESSES__SHORTED: list[str] = []
    ADDRESSES__PAIRED: list[tuple[str, str]] = []

    # CMDS -----------------------------------------------------
    CMD__RESET: str = "RST"

    # INIT ============================================================================================================
    pass
    pass
    pass
    pass
    pass
    pass
    pass

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._SERIAL = Serial()
        self.history = HistoryIO()

        # apply settings
        # self._SERIAL.interCharTimeout = 0.8
        self._SERIAL.baudrate = self.BAUDRATE
        self._SERIAL.timeout = self.TIMEOUT__READ
        # self._SERIAL.write_timeout = self._TIMEOUT__WRITE

        # self.addresses_system__detect()   # DONT USE in init!!!
        # self.addresses_shorted__detect()   # DONT USE in init!!!
        # self.addresses_paired__detect()   # DONT USE in init!!!

    def __del__(self):
        self._address__release()
        self.disconnect()

    def cmd_prefix__set(self) -> None:
        """
        OVERWRITE IF NEED/USED!
        """
        # self.PREFIX = ""
        return

    # CONNECT =========================================================================================================
    pass
    pass
    pass
    pass
    pass
    pass
    pass

    def check__connected(self) -> bool:
        try:
            return self._SERIAL.is_open
        except:
            return False

    def disconnect(self, release: bool = None) -> None:
        """
        for release - its better to use direct cls._addresses__release()
        """
        if release or self.RELEASE_ON_DISCONNECT:
            self._address__release()

        try:
            self._SERIAL.close()
        except:
            pass

        try:
            self._EMULATOR__INST.disconnect()
        except:
            pass

    def connect__only_if_address_resolved(self) -> Union[None, bool, NoReturn]:
        if self.address_check__resolved():
            return self.connect()

    def connect(
            self,
            address: TYPE__ADDRESS | None = None,
            _raise: bool | None = None,
            _touch_connection: bool | None = None
            # no final connection! specially keep ability to connect without Emu on cls main perpose (search ports)!
    ) -> Union[bool, NoReturn]:
        msg = None
        exx = None
        need_open = True

        # SETTINGS ---------------------------------
        if _raise is None:
            _raise = self.RAISE_CONNECT

        if address is None:
            address = self.ADDRESS

        if address in Type__AddressAutoAcceptVariant:
            if not self.address__resolve(address):
                exx = Exx_SerialAddress_NotApplyed()
                need_open = False
            else:
                address = self.ADDRESS

        # need_open ==========================================================
        # CHANGE PORT OR USE SAME ---------------------------------
        if need_open:
            if self._SERIAL.port != address:
                # close old
                if self._SERIAL.is_open:
                    self._SERIAL.close()

                # set new
                self._SERIAL.port = address
                if self._SERIAL.is_open:
                    self._SERIAL.port = None
                    msg = f"[ERROR] Attempt to connect to already opened port IN OTHER OBJECT {self._SERIAL}"
                    exx = Exx_SerialAddress_AlreadyOpened_InOtherObject(msg)

                    need_open = False
            else:
                if self._SERIAL.is_open:
                    need_open = False

        # Try OPEN ===================================================================
        if need_open:
            try:
                self._SERIAL.open()
                # self.write_eol()
                # self.buffers_clear()
            except Exception as _exx:
                if not _touch_connection:
                    self.LOGGER.error(f"[{self._SERIAL.port}]{_exx!r}")

                if "FileNotFoundError" in str(_exx):
                    msg = f"[ERROR] PORT NOT EXISTS IN SYSTEM {self._SERIAL}"
                    exx = Exx_SerialAddress_NotExists(repr(_exx))

                    # self.detect_available_ports()

                elif "Port must be configured before" in str(_exx):
                    msg = f"[ERROR] PORT NOT CONFIGURED {self._SERIAL}"
                    exx = Exx_SerialAddress_NotApplyed(repr(_exx))

                elif "PermissionError" in str(_exx):
                    msg = f"[ERROR] PORT ALREADY OPENED {self._SERIAL}"
                    exx = Exx_SerialAddress_AlreadyOpened(repr(_exx))

                else:
                    msg = f"[ERROR] PORT OTHER ERROR {self._SERIAL}"
                    exx = Exx_SerialAddress_OtherError(repr(_exx))

        # FINISH -----------------------------------------------
        # FAIL -----------------------------
        if exx:
            if not _touch_connection:
                self.LOGGER.error(f"[{self._SERIAL.port}]{msg}")

            if _raise:
                raise exx
            else:
                return False

        # OK -----------------------------
        if not _touch_connection:
            if not self.connect__validate():
                self.disconnect()
                return False

            self._address__occupy(address)
            self.emulator_start()
            self.cmd_prefix__set()

        return True

    def connect__validate(self) -> bool:
        """
        DIFFERENCE
        ----------
        connect validation used always!
        address validation used only in step of detecting address (after execution address would be set to exact string value)
        """
        return True

    def emulator_start(self) -> None:
        if not self._EMULATOR__START:
            return

        if self._EMULATOR__CLS:
            self._EMULATOR__INST = self._EMULATOR__CLS(self.address_paired__get())

        if self._EMULATOR__INST.connect():
            self._EMULATOR__INST.start()
            self._EMULATOR__INST.wait__cycle_active()
            self._buffers_clear__read()

    # ADDRESS =========================================================================================================
    """
    THIS IS USED for applying by SIMPLE WAY just exact address! 
    """
    pass
    pass
    pass
    pass
    pass
    pass
    pass

    @property
    def ADDRESS(self) -> TYPE__ADDRESS:
        return self._ADDRESS

    @ADDRESS.setter
    def ADDRESS(self, value: TYPE__ADDRESS) -> None:
        if value == self.ADDRESS:
            return

        self._address__release()
        # self.disconnect()     # dont place here! incorrect logic!
        self._ADDRESS = value
        self._address__occupy(value)

    @classmethod
    @property
    def ADDRESSES__FREE(cls) -> list[str]:
        cls.addresses_system__detect()
        result = []
        for port_name, port_owner in cls.ADDRESSES__SYSTEM.items():
            if port_owner is None:
                result.append(port_name)
        return result

    # OCCUPATION ------------------------------------------------------------------------------------------------------
    def _address__occupy(self, address: TYPE__ADDRESS | None = None) -> None:
        """
        USAGE
        -----
        usually when start module we have exact and only one set of UartDevices!
        ones we connect exact device - it would occupy the port.
        port must be occupied even if Serial is not open!
        """
        if address is None:
            address = self.ADDRESS

        if address in SerialClient.ADDRESSES__SYSTEM:
            SerialClient.ADDRESSES__SYSTEM[address] = self

        self.LOGGER.info(f"[{self._SERIAL.port}][OK] connected/locked/occupy {self._SERIAL}")

    def _address__release(self) -> None:
        """
        USE ONLY when change address!
        """
        if self in SerialClient.ADDRESSES__SYSTEM.values():
            for address, owner in SerialClient.ADDRESSES__SYSTEM.items():
                if owner is self:
                    SerialClient.ADDRESSES__SYSTEM[address] = None
                    # self.disconnect()
                    break

    @classmethod
    def _addresses__release(cls) -> None:
        """
        make all ports vacant for autodetect.
        use direct Clear() if need reinit items.

        CREATED SPECIALLY FOR
        ---------------------
        test issues (testing this module!)
        when set of tests is finished for one connection schema - release all
        """
        for key in cls.ADDRESSES__SYSTEM:
            cls.ADDRESSES__SYSTEM[key] = None

    def address_forget(self) -> None:
        """
        GOAL
        ----
        get default address for derivatives or else just clear to None.

        SPECIAL CREATED FOR
        -------------------
        testplans module - to drop all devices for manually refind devices

        CONSTRAINTS
        -----------
        be sure to use Base class from Derivatives so method would work correctly (as designed)!
        """
        self.ADDRESS = None

        # CAREFUL ---- here we will get curriculum import!
        # from . import SerialClient_FirstFree, SerialClient_FirstFree_Shorted, SerialClient_FirstFree_Paired, SerialClient_FirstFree_AnswerValid
        # if isinstance(self, SerialClient_FirstFree):
        #     self.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE
        # elif isinstance(self, SerialClient_FirstFree_Shorted):
        #     self.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__SHORTED
        # elif isinstance(self, SerialClient_FirstFree_Paired):
        #     self.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__PAIRED
        # elif isinstance(self, SerialClient_FirstFree_AnswerValid):
        #     self.ADDRESS = Type__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID

    def address__resolve(self, address: TYPE__ADDRESS | None = None) -> bool:
        """
        GOAL
        ----
        resolve address passed as Type__AddressAutoAcceptVariant
        could be call separately! before connect!

        SPECIALLY CREATED FOR
        ---------------------
        separate address resolve and connect

        NOTE
        ----
        :param address: if result if True - ALWAYS APPLY final value in instance!
        """
        if address is None:
            address = self.ADDRESS

        # RESOLVE ---------------------------------
        if address in Type__AddressAutoAcceptVariant:
            if address == Type__AddressAutoAcceptVariant.FIRST_FREE:
                address = self.address_get__first_free()
            elif address == Type__AddressAutoAcceptVariant.FIRST_FREE__SHORTED:
                address = self.address_get__first_free__shorted()
            elif address == Type__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID:
                address = self.address_get__first_free__valid()
            elif address == Type__AddressAutoAcceptVariant.FIRST_FREE__PAIRED:
                address = self.address_get__first_free__paired()

        # APPLY -----------------------------------
        result = isinstance(address, str)
        if result:
            self.ADDRESS = address
        return result

    def address_check__resolved(self, address: TYPE__ADDRESS | None = None) -> bool:
        if address is None:
            address = self.ADDRESS
        return isinstance(address, str)

    def address_check__occupied(self, address: TYPE__ADDRESS | None = None) -> bool:
        if not self.address_check__resolved(address):
            return False

        if address is None:
            address = self.ADDRESS
        owner = self.ADDRESSES__SYSTEM.get(address)
        return owner is not None

    # AUTODETECT ------------------------------------------------------------------------------------------------------
    pass  # dont move this all to CLASSMETHOD!!!

    def address_get__first_free(self) -> str | None:
        result = None
        for address in self.ADDRESSES__FREE:
            if self.connect(address=address, _raise=False, _touch_connection=True):
                result = address

            self.disconnect()
            if result:
                break

        msg = f"[{result=}]_address_apply__first_free"
        if result:
            self.LOGGER.info(msg)
        else:
            self.LOGGER.warning(msg)

        return result

    def address_get__first_free__shorted(self) -> str | None:
        """
        dont overwrite! dont mess with address__autodetect_logic!
        used to find exact device in all comport by some special logic like IDN/NAME value
        """
        result = None
        for address in self.addresses_shorted__detect():
            if self.address_check__occupied(address):
                continue

            if self.connect(address=address, _raise=False, _touch_connection=True):
                result = address

            self.disconnect()
            if result:
                break

        msg = f"[{result=}]_address_apply__first_free__shorted"
        if result:
            self.LOGGER.info(msg)
        else:
            self.LOGGER.warning(msg)

        return result

    def address_get__first_free__valid(self) -> str | None:
        """
        dont overwrite! dont mess with address__autodetect_logic!
        used to find exact device in all comport by some special logic like IDN/NAME value
        """
        result = None
        for address in self.ADDRESSES__FREE:
            if self.connect(address=address, _raise=False, _touch_connection=True):
                try:
                    if self.address__validate():
                        result = address
                except Exception as exx:
                    print(f"finding address {exx!r}")
                    pass

                self.disconnect()
                if result:
                    break

        msg = f"[{result=}]_address_apply__first_free__answer_valid"
        if result:
            self.LOGGER.info(msg)
        else:
            self.LOGGER.warning(msg)

        return result

    def address_get__first_free__paired(self) -> str | None:
        """
        connect only for first address!
        need for occupy addresses by several servers before starting main process!!!
        in case of any address we could occupy by two servers whole pair!

        secondary address you should get by special methods for pair or by address_validating

        cls.pairs = {
            0: (COM1, COM2),
            1: (COM3, COM4),
        }

        cls.pairs = {
            "ATC": (COM1, COM2),    # change name
            1: (COM3, COM4),
        }
        """
        result = None
        for address, _ in self.addresses_paired__detect():
            if self.connect(address=address, _raise=False, _touch_connection=True):
                result = address

            self.disconnect()
            if result:
                break

        # FINISH -------------
        msg = f"[{result=}]address_get__first_free__paired"
        if result:
            self.LOGGER.info(msg)
        else:
            self.LOGGER.warning(msg)

        return result

    # VALIDATION ------------------------------------------------------------------------------------------------------
    def address__validate(self) -> bool | None | NoReturn:
        """
        overwrite for you case!
        used to find exact device in all comport by some special logic like IDN/NAME value.

        IDEA:
        1. this func will exec on every accessible address
        2. if this func return True - address would be accepted!
        3. raiseExx/NoReturn - equivalent as False!
        """

    def _address__validate_shorted(self) -> bool | None:
        """
        this is the internal method! for autodetect shorted address
        """
        load = "FIND__SHORTED"
        return self.write_read__last_validate(load, load)

    # DETECT ----------------------------------------------------------------------------------------------------------
    @staticmethod
    def address__check_exists(address: str) -> bool:
        try:
            inst = SerialClient()
            inst.connect(address=address, _raise=True, _touch_connection=True)
            inst.disconnect()
        except Exx_SerialAddress_NotExists:
            return False
        except:
            return False
        return True

    @classmethod
    def addresses_system__detect(cls) -> dict[str, Union[None, Self]]:
        if SerialClient.ADDRESSES__SYSTEM:
            return SerialClient.ADDRESSES__SYSTEM

        # WORK -------------------------------------------------------
        result = cls._addresses_system__detect_1__standard_method()
        for port in cls._addresses_system__detect_2__direct_access():
            if port not in result:
                result.append(port)

        # result -------------------------------------------------------
        if result:
            print(f"[OK] detected serial ports {result}")
        else:
            print("[WARN] detected no serial ports")

        SerialClient.ADDRESSES__SYSTEM.clear()
        SerialClient.ADDRESSES__SYSTEM.update(**dict.fromkeys(result, None))
        return SerialClient.ADDRESSES__SYSTEM

    @staticmethod
    def _addresses_system__detect_1__standard_method() -> Union[list[str], NoReturn]:
        """
        WINDOWS - USB
            ==========OBJECTINFO.PRINT==========================================================================
            str=COM8 - PL2303HXA PHASED OUT SINCE 2012. PLEASE CONTACT YOUR SUPPLIER.
            repr=<serial.tools.list_ports_common.ListPortInfo object at 0x00000267B3A38200>
            ----------properties_ok-----------------------------------------------------------------------------
            description              	str       :COM8
            device                   	str       :COM8
            hwid                     	str       :USB VID:PID=067B:2303 SER= LOCATION=1-2.4.3
            interface                	NoneType  :None
            location                 	str       :1-2.4.3
            manufacturer             	str       :Prolific
            name                     	str       :COM8
            pid                      	int       :8963
            product                  	NoneType  :None
            serial_number            	str       :
            vid                      	int       :1659
            ----------properties_exx----------------------------------------------------------------------------
            ----------aux_types-----------------------------------------------------------------------------------
            ----------methods_ok--------------------------------------------------------------------------------
            apply_usb_info           	NoneType  :None
            usb_description          	str       :COM8
            usb_info                 	str       :USB VID:PID=067B:2303 SER= LOCATION=1-2.4.3
            ----------methods_exx-------------------------------------------------------------------------------
            ====================================================================================================

        RASPBERRY - USB
        кажется не видит встроенный COM порт даже после его включения и перезагрузки - только USB!!!

            ==========OBJECTINFO.PRINT==========================================================================
            str=/dev/ttyUSB0 - USB-Serial Controller
            repr=<serial.tools.list_ports_linux.SysFS object at 0x7fb332d9d0>
            ----------properties_ok-----------------------------------------------------------------------------
            description                     str       :USB-Serial Controller
            device                          str       :/dev/ttyUSB0
            device_path                     str       :/sys/devices/platform/soc/3f980000.usb/usb1/1-1/1-1.4/1-1.4:1.0/ttyUSB0
            hwid                            str       :USB VID:PID=067B:2303 LOCATION=1-1.4
            interface                       NoneType  :None
            location                        str       :1-1.4
            manufacturer                    str       :Prolific Technology Inc.
            name                            str       :ttyUSB0
            pid                             int       :8963
            product                         str       :USB-Serial Controller
            serial_number                   NoneType  :None
            subsystem                       str       :usb-serial
            usb_device_path                 str       :/sys/devices/platform/soc/3f980000.usb/usb1/1-1/1-1.4
            usb_interface_path              str       :/sys/devices/platform/soc/3f980000.usb/usb1/1-1/1-1.4/1-1.4:1.0
            vid                             int       :1659
            ----------properties_exx----------------------------------------------------------------------------
            ----------aux_types-----------------------------------------------------------------------------------
            ----------methods_ok--------------------------------------------------------------------------------
            apply_usb_info                  NoneType  :None
            usb_description                 str       :USB-Serial Controller
            usb_info                        str       :USB VID:PID=067B:2303 LOCATION=1-1.4
            ----------methods_exx-------------------------------------------------------------------------------
            read_line                       TypeError :join() missing 1 required positional argument: 'a'
            ====================================================================================================
        """
        result: list[str] = []

        # find not opened ports ----------------------------------------
        for obj in list_ports.comports():
            # ObjectInfo(obj).print(hide_skipped=True, hide_build_in=True)
            result.append(obj.device)
            if Exx_SerialPL2303IncorrectDriver.MARKER in str(obj):
                msg = f'[WARN] incorrect driver [{Exx_SerialPL2303IncorrectDriver.MARKER}]'
                print(msg)
                raise Exx_SerialPL2303IncorrectDriver(msg)
        return result

    @staticmethod
    def _addresses_system__detect_2__direct_access() -> Union[list[str], NoReturn]:
        """Определяет список портов (НЕОТКРЫТЫХ+ОТКРЫТЫХ!!!) - способом 2 (слепым тестом подключения и анализом исключений)
        Всегда срабатывает!
        """
        if sys.platform.startswith('win'):
            attempt_list = ['COM%s' % (i + 1) for i in range(256)]
        elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
            # this excludes your current terminal "/dev/tty"
            attempt_list = glob.glob('/dev/tty[A-Za-z]*')
        elif sys.platform.startswith('darwin'):
            attempt_list = glob.glob('/dev/tty.*')
        else:
            raise EnvironmentError('[ERROR]Unsupported platform')

        result: list[str] = []
        for name in attempt_list:
            if SerialClient.address__check_exists(address=name):
                result.append(name)

        return result

    @classmethod
    def addresses_shorted__detect(cls) -> list[str]:
        if SerialClient.ADDRESSES__SHORTED:
            return SerialClient.ADDRESSES__SHORTED

        # WORK ----------------------------------------------
        result = []
        for address in cls.addresses_system__detect():
            obj = SerialClient()
            if obj.connect(address=address, _raise=False, _touch_connection=True):
                if obj._address__validate_shorted():
                    result.append(address)
                obj.disconnect()

        SerialClient.ADDRESSES__SHORTED = result
        print(f"{SerialClient.ADDRESSES__SHORTED=}")
        return result

    @classmethod
    def addresses_paired__detect(cls) -> list[tuple[str, str]]:
        # FIXME: IT SEEMS NOT WORKING!!!
        if cls.addresses_system__count() < 2:
            return []

        if SerialClient.ADDRESSES__PAIRED:
            return SerialClient.ADDRESSES__PAIRED

        # WORK ----------------------------------------------
        load = "EXPECT_ANSWER__PAIRED"
        result = []
        instances_free_remain = []

        system_ports = cls.addresses_system__detect()

        # CONNECT ALL -----
        for address in system_ports:
            instance = SerialClient()
            if instance.connect(address=address, _raise=False, _touch_connection=True):
                instances_free_remain.append(instance)

        while len(instances_free_remain) > 1:
            # take one port --------
            master = instances_free_remain.pop(0)
            _write_success = master._write(load)
            master.disconnect()
            time.sleep(0.5)

            # try found pair ------
            for index, slave in enumerate(instances_free_remain):
                read_lines = slave.read_lines(_timeout=2)
                if read_lines:
                    if read_lines[-1] == load or list(read_lines)[-1] == load:
                        result.append((master._SERIAL.port, slave._SERIAL.port))
                        instances_free_remain.pop(index)
                        slave.disconnect()
                        break

        # disconnect all remain ---------
        for remain in instances_free_remain:
            remain.disconnect()

        SerialClient.ADDRESSES__PAIRED = result
        print(f"{SerialClient.ADDRESSES__PAIRED=}")
        return result

    def addresses_paired__get_used(self) -> Optional[tuple[str, str]]:
        for pair in self.addresses_paired__detect():
            if self._SERIAL.port in pair:
                return pair

    def address_paired__get(self) -> str | None:
        if not SerialClient.ADDRESSES__PAIRED:
            SerialClient.addresses_paired__detect()

        if not self.address_check__resolved():
            return

        for pair in SerialClient.ADDRESSES__PAIRED:
            if self.ADDRESS in pair:
                addr1, addr2 = pair
                if self.ADDRESS == addr1:
                    return addr2
                else:
                    return addr1

    # COUNTS -----------------------------------------
    @classmethod
    def addresses_system__count(cls) -> int:
        return len(cls.addresses_system__detect())

    @classmethod
    def addresses_shorted__count(cls) -> int:
        return len(cls.addresses_shorted__detect())

    @classmethod
    def addresses_paired__count(cls) -> int:
        return len(cls.addresses_paired__detect())

    @classmethod
    def addresses_free__count(cls) -> int:
        return len(cls.ADDRESSES__FREE)

    # RW ==============================================================================================================
    pass
    pass
    pass
    pass
    pass
    pass
    pass

    # BUFFERS ---------------------------------------------------------------------------------------------------------
    def _buffers_clear__read(self, _timeout: Optional[float] = None) -> None:
        self._SERIAL.timeout = _timeout or self.TIMEOUT__READ or None
        while True:
            try:
                if not self._SERIAL.readline(1):
                    break
            except:
                pass
            time.sleep(0.1)

        self._SERIAL.timeout = self.TIMEOUT__READ or None  # set back - final

    def _buffers_clear__write(self) -> None:
        """useful to drop old previous incorrect send msg! in other words it is clear/reinit write buffer!

        here we need just finish line by correct/exact EOL if it was previously send without it or with incorrect.
        """
        try:
            self.write_eol()
        except:
            pass

    def buffers_clear(self) -> None:
        """useful to drop old previous incorrect send msg! in other words it is clear/reinit write buffer!

        here we need just finish line by correct/exact EOL if it was previously send without it or with incorrect.
        """
        self._buffers_clear__write()
        self._buffers_clear__read()

    # CMD -------------------------------------------------------------------------------------------------------------
    def _create_cmd_line(self, cmd: str, prefix: Optional[str] = None, args: list[Any] = None,
                         kwargs: dict[Any, Any] = None) -> str:
        result = ""

        cmd = self._data_ensure__string(cmd)
        cmd = self._data_eol__clear(cmd)

        if prefix is None:
            prefix = self.PREFIX or ""

        if prefix and not cmd.startswith(prefix):
            result += f"{prefix}"

        result += f"{cmd}"

        if args:
            for arg in args:
                result += f" {arg}"

        if kwargs:
            for key, value in kwargs.items():
                result += f" {key}={value}"
        return result

    # SUCCESS ---------------------------------------------------------------------------------------------------------
    def answer_is_success(self, data: AnyStr) -> bool:
        data = self._data_ensure__string(data)
        return self.ANSWER_SUCCESS.upper() == data.upper()

    def answer_is_fail(self, data: AnyStr, _raise: Optional[bool] = None) -> Union[bool, NoReturn]:
        if _raise is None:
            _raise = self.RAISE_READ_FAIL_PATTERN

        data = self._data_ensure__string(data)
        if isinstance(self.ANSWER_FAIL_PATTERN, str):
            patterns = [self.ANSWER_FAIL_PATTERN, ]
        else:
            patterns = self.ANSWER_FAIL_PATTERN

        for pattern in patterns:
            if re.search(pattern, data, flags=re.IGNORECASE):
                if _raise:
                    msg = f"[ERROR] match fail [{pattern=}/{data=}]"
                    self.LOGGER.error(f"[{self._SERIAL.port}]{msg}")
                    raise Exx_SerialRead_FailPattern(msg)
                else:
                    return True

        return False

    # BYTES -----------------------------------------------------------------------------------------------------------
    @classmethod
    def _bytes_eol__ensure(cls, data: bytes) -> Union[bytes, NoReturn]:
        data = cls._data_eol__clear(data) + cls.EOL__SEND
        return data

    @classmethod
    def _data_eol__clear(cls, data: AnyStr) -> Union[AnyStr, NoReturn]:
        result: bytes = cls._data_ensure__bytes(data)
        while True:
            data_new = result.strip()
            data_new = data_new.strip(cls.EOL__UNI_SET)
            if not result or result == data_new:
                break
            result = data_new

        if isinstance(data, str):
            return cls._data_ensure__string(result)

        return result

    @classmethod
    def _bytes_edition__apply(cls, data: bytes) -> bytes:
        """not used so far!!!
        need to handle editions used by hand in manual typing from terminal!
        """
        # TODO: finish or leave blank! its not so necessary!!! and seems a little bit hard to apply
        return data

    # BYTES/STR -------------------------------------------------------------------------------------------------------
    @classmethod
    def _data_ensure__bytes(cls, data: AnyStr) -> bytes:
        if isinstance(data, bytes):
            return data
        else:
            return bytes(data, encoding=cls.ENCODING)

    @classmethod
    def _data_ensure__string(cls, data: AnyStr) -> Union[str, NoReturn]:
        """
        EXCEPTION ORIGINAL VARIANT
        --------------------------
            cls = <class 'test__serial_server.Victim'>, data = b'\x85RR__NAME_CMD_OR_PARAM'

                @classmethod
                def _data_ensure__string(cls, data: AnyStr) -> Union[str, NoReturn]:
                    if isinstance(data, bytes):
            >           return data.decode(encoding=cls.ENCODING)
            E           UnicodeDecodeError: 'utf-8' codec can't decode byte 0x85 in position 0: invalid start byte

            buses\serial_client.py:722: UnicodeDecodeError
        """
        # TODO: move to sepaarted module!
        try:
            if isinstance(data, bytes):
                result = data.decode(encoding=cls.ENCODING)
            else:
                result = str(data)
        except Exception as exx:
            print(f"{exx!r}")
            msg = f"[FAIL] decoding {data=}"
            raise Exx_SerialRead_FailDecoding(msg)

        # check printable ascii ------------------
        for char in result:
            if isinstance(char, bytes):
                char = chr(char)
            if char not in string.printable:
                msg = f"[FAIL] decoding [{char=}]"
                raise Exx_SerialRead_FailDecoding(msg)
        # ----------------------------------------

        return result

    # R ---------------------------------------------------------------------------------------------------------------
    # TODO: use wrapper for connect/disconnect!??? - NO!
    def read_lines(self, _timeout: Optional[float] = None) -> list[TYPE__RW_ANSWER_SINGLE] | NoReturn:
        result: list[str] = []
        while True:
            line = self.read_line(_timeout)
            _timeout = self.TIMEOUT__READ or None
            if line is None or str(line) == "":
                break
            else:
                result.append(line)

        return result

    def read_line(self, _timeout: Optional[float] = None) -> Union[str, ValueUnit, NoReturn]:
        """
        read line from bus buffer,

        CAREFUL
        -------
        DONT USE! IF WANT TO BE SURE!
        instead use write_read__last_validate!!! it would rewrite data if not valid answer (even with incorrect but good decoding)!!!

        TODO: here possible rewrite by getting from history!!!
        """
        # FIXME: return Object??? need keep exx for not finished readline!!!
        # var1: just read as usual - could cause error with not full bytes read in ONE CHAR!!!
        # data = self._SERIAL.readline()

        # var2: char by char
        data = b""
        eol_received = False

        self._SERIAL.timeout = _timeout or self.TIMEOUT__READ or None
        while True:
            new_char = None
            for i in range(2):
                new_char = self._SERIAL.readline(1)
                self._SERIAL.timeout = self.TIMEOUT__READ or None  # set back - no need repiatiting
                if new_char:
                    break

            if not new_char:
                # print(f"detected finish line")
                break
            if new_char == b'\x7f':  # BACKSPACE
                # LINE EDITION --------------
                data = data[:-1]
                continue

            if new_char in self.EOL__UNI_SET:
                # LINE FINISHED --------------
                eol_received = True
                if data:
                    break
                else:
                    continue

            data += new_char

        self._SERIAL.timeout = self.TIMEOUT__READ or None  # set back - final

        # RESULT ----------------------
        if data:
            if not eol_received:
                msg = f"[ERROR]NotFullLine read_line={data}->CLEAR!!!"
                exx = Exx_SerialRead_NotFullLine(msg)
                data = b""
            else:
                msg = f"[OK]read_line={data}"

        else:
            msg = f"[WARN]BLANK read_line={data}"

        self.LOGGER.info(f"[{self._SERIAL.port}]{msg}")

        data = self._bytes_edition__apply(data)
        data = self._data_eol__clear(data)
        data = self._data_ensure__string(data)
        self.history.add_output(data)
        self.answer_is_fail(data)

        # NOTE: dont delete! need to direct cmp
        try:
            # pass
            data = ValueUnit_NoMulty(data)
        except:
            pass

        return data

    # W ---------------------------------------------------------------------------------------------------------------
    def _write(
            self,
            data: Union[AnyStr, list[AnyStr]],
            prefix: Optional[str] = None,
            args: Optional[list] = None,
            kwargs: Optional[dict] = None
    ) -> bool:
        """
        just send data into bus!
        usually we dont need just send without reading! so it useful for debugging

        :return: result of sent

        args/kwargs - used only for single line!!!
        """
        args = args or []
        kwargs = kwargs or {}

        data = data or ""

        # LIST -----------------------
        if isinstance(data, (list, tuple,)):
            if len(data) > 1:
                for data_i in data:
                    if not self._write(data=data_i, prefix=prefix, args=args, kwargs=kwargs):
                        return False
                return True
            else:
                data = data[0]

        # SINGLE ---------------------
        data = self._create_cmd_line(cmd=data, prefix=prefix, args=args, kwargs=kwargs)
        self.history.add_input(self._data_ensure__string(data))

        data = self._data_ensure__bytes(data)
        data = self._bytes_eol__ensure(data)

        try:
            data_length = self._SERIAL.write(data)
            msg = f"[OK]write={data}/{data_length=}"
        except Exception as exx:
            msg = f"[FAIL]write={data}/{exx!r}"
            self.LOGGER.warning(f"[{self._SERIAL.port}]{msg}")
            return False

        if data_length > 0:
            self.LOGGER.info(f"[{self._SERIAL.port}]{msg}")
            return True
        else:
            msg = f"[ERROR] write {data}"
            self.LOGGER.error(f"[{self._SERIAL.port}]{msg}")
            return False

    def write_eol(self, eol: bytes = None) -> bool:
        """
        write without any conditions (no prefixes! no suffixes)!
        GOAL
        ----
        finish any writen cmd and clear buffer
        """
        eol = eol or self.EOL__SEND
        try:
            return bool(self._SERIAL.write(eol))
        except:
            return False

    def write_read(
            self,
            data: Union[AnyStr, list[AnyStr]],
            prefix: Optional[str] = None,
            args: Optional[list] = None,
            kwargs: Optional[dict] = None,

            retry_noanswer: int | None = None,
            _timeout: Optional[float] = None,
    ) -> Union[HistoryIO, NoReturn]:
        """
        send data and return history

        CAREFUL
        -------
        in this case you can get incorrect but printable chars!!!
        DONT USE! IF WANT TO BE SURE!
        instead use write_read__last_validate!!! it would rewrite data if not valid answer (even with incorrect but good decoding)!!!
        """
        history = HistoryIO()
        if retry_noanswer is None:
            retry_noanswer = self.REWRITEIF_READNOANSWER
        remain__retry_noanswer = retry_noanswer or 0

        # LIST -------------------------
        if isinstance(data, (list, tuple,)):
            for data_i in data:
                history_i = self.write_read(data_i, prefix=prefix, args=args, kwargs=kwargs,
                                            retry_noanswer=retry_noanswer)
                history.add_history(history_i)
        else:
            # SINGLE LAST -----------------------
            data_o = []
            remain__retry_decode = self.REWRITEIF_READFAILDECODE or 0
            while remain__retry_decode >= 0 and remain__retry_noanswer >= 0:
                if self._write(data=data, prefix=prefix, args=args, kwargs=kwargs):
                    try:
                        data_o = self.read_lines(_timeout=_timeout)
                        if data_o:  # here are validated string data!   # NEED ALWAYS GET ANYTHING IN RESPONSE
                            break
                        else:
                            remain__retry_noanswer -= 1

                    except Exx_SerialRead_FailDecoding:
                        remain__retry_decode -= 1

                    self.buffers_clear()
            history.add_io(self._data_ensure__string(data), data_o)

        # RESULT ----------------------------
        return history

    def write_read__last(
            self,
            data: Union[AnyStr, list[AnyStr]],
            prefix: Optional[str] = None,
            args: Optional[list] = None,
            kwargs: Optional[dict] = None,

            retry_noanswer: int | None = None,
            _timeout: Optional[float] = None,
    ) -> Union[str, NoReturn]:
        """
        it is created specially for single cmd usage! but decided leave multy cmd usage as feature.
        return always last_output

        CAREFUL
        -------
        in this case you can get incorrect but printable chars!!!
        DONT USE! IF WANT TO BE SURE!
        instead use write_read__last_validate!!! it would rewrite data if not valid answer (even with incorrect but good decoding)!!!
        """
        return self.write_read(
            data=data,
            prefix=prefix,
            args=args,
            kwargs=kwargs,
            retry_noanswer=retry_noanswer,
            _timeout=_timeout,
        ).last_output

    def write_read__last_validate(
            self,
            input: Union[str, list[str]] | None,
            expect: Union[Any, list[Any]],
            prefix: Optional[str] = None,
            args: Optional[list] = None,
            kwargs: Optional[dict] = None,

            retry_novalid: int | None = None,
            _timeout: Optional[float] = None,
            _as_regexp: Optional[bool] = None,
    ) -> bool:
        """
        CREATED SPECIALLY FOR
        ---------------------
        1. address__validate

        SOLVE PROBLEMS
        --------------
        communicate with wrong working devices!

        :param expect: can be Any! even Valid/ValueUbit/ValueVariants as validation object!
        """
        if retry_novalid is None:
            retry_novalid = self.REWRITEIF_NOVALID or 0

        if isinstance(expect, (tuple, list, set, dict)):
            expect_list = expect
        else:
            expect_list = [expect, ]

        while retry_novalid >= 0:
            if input:
                output_last = self.write_read__last(
                    data=input,
                    prefix=prefix,
                    args=args,
                    kwargs=kwargs,
                    retry_noanswer=0,  # use only retry_noanswer=0! so it would not multiplyed iterations
                    _timeout=_timeout,
                )
            else:
                outputs = self.read_lines(_timeout=_timeout)
                if outputs:
                    output_last = outputs[-1]
                else:
                    output_last = ""

            for expect_var in expect_list:
                if _as_regexp:
                    if re.fullmatch(expect_var, str(output_last), flags=re.IGNORECASE):
                        return True
                else:
                    if (
                            str(output_last).lower() == str(expect_var).lower()
                            or
                            EqAux(output_last).check_doubleside__bool(expect_var)
                    ):
                        return True

            retry_novalid -= 1
            print(f"{retry_novalid=}")

        return False

    def write_read__last_validate_regexp(self, *args, **kwargs) -> bool:
        """
        created specially for address__validate
        """
        return self.write_read__last_validate(*args, **kwargs, _as_regexp=True)

    def dump_cmds(self, cmds: list[str] = None) -> Union[HistoryIO, NoReturn]:
        """
        useful to get results for standard cmds list
        if you need to read all params from device!
        """
        cmds = cmds or self.CMDS_DUMP
        history = self.write_read(cmds)
        history.print_io()
        return history

    # =================================================================================================================
    def __getattr__(self, item: str) -> Callable[..., Union[str, NoReturn]]:
        """if no exists attr/meth

        USAGE COMMANDS MAP
        ==================

        1. SHOW (optional) COMMANDS EXPLICITLY by annotations without values!
        -----------------------------------------------------------------
            class MySerialDevice(SerialClient):
                IDN: Callable[[Any], TYPE__RW_ANSWER]
                ADDR: Callable[[Any], TYPE__RW_ANSWER]
                DUMP: Callable[[Any], TYPE__RW_ANSWER]

        2. USE in code
        --------------
            dev = MySerialDevice()
            dev.connect()
            dev.IDN()       --> dev.write_read__last("IDN")
            dev.VIN()       # dev.write_read__last("VIN")
            dev.VIN(12)     # dev.write_read__last("VIN 12")
            dev.VIN("12")   # dev.write_read__last("VIN 12")
            dev.VIN("12 13")  # dev.write_read__last("VIN 12 13")
            dev.VIN(12, 13)   # dev.write_read__last("VIN 12 13")
            dev.VIN(CH1=12, CH2=13) # dev.write_read__last("VIN CH1=12 CH2=13")
            dev.VIN(12, CH2=13)     # dev.write_read__last("VIN 12 CH2=13") by args/kwargs

            # ALL VARIANTS
            dev.VIN(11, CH2=13)
            dev.VIN__11(22, CH2=13)
            dev.send__VIN(11, CH2=13)
            dev.send__VIN__11(22, CH2=13)

            # KWARGS for WriteRead method
            dev.TEST("HELLO", CH2=13, __timeout=5) --> dev.write_read__last("TEST CH2=13", _timeout=5) # note underscore

        """
        item_args = []  # args in getattr name

        # 1=apply _GETATTR_STARTSWITH__SEND
        if self._GETATTR_STARTSWITH__SEND and item.startswith(self._GETATTR_STARTSWITH__SEND):
            item = item.replace(self._GETATTR_STARTSWITH__SEND, "")

        # 2=apply args in getattrs name
        if "__" in item:
            item_splited = item.split(self._GETATTR_SPLITTER__ARGS)
            item = item_splited[0]
            item_args = item_splited[1:]

        # 3=separate cmd/meth kwargs
        # cand to it in here!

        # 4=apply direct cmd
        result = lambda *_args, **_kwargs: self.write_read__last(
            data=self._create_cmd_line(
                cmd=item,
                args=[*item_args, *_args],
                kwargs={k: v for k, v in _kwargs.items() if not k.startswith("_")}
            ),
            **{k[1:]: v for k, v in _kwargs.items() if k.startswith("_")}
        )
        return result

    # =================================================================================================================
    # USER COMMANDS
    # -----------------------------------------------------------------------------------------------------------------
    def reset(self, sleep_after: float = 1) -> None:
        if self.connect__only_if_address_resolved():
            self.write_read(self.CMD__RESET)
            self._buffers_clear__read(sleep_after)
        pass


# =====================================================================================================================



import time

from base_aux.testplans.tc import *
from base_aux.valid.m2_valid_base2_derivatives import *

from .tc0_groups import *


# =====================================================================================================================
class TestCase(TcGroup_ATC220220, TestCaseBase):
    ASYNC = True
    DESCRIPTION = "atc"

    @classmethod
    def startup__cls__wrapped(cls) -> TYPE__RESULT_W_NORETURN:
        return True
        result_chain = ValidChains(
            [
                Valid(value_link=hasattr(cls, "DEVICES__BREEDER_CLS"), name="hasattr DEVICES__CLS"),
                Valid(value_link=hasattr(cls.DEVICES__BREEDER_CLS, "ATC"), name="hasattr ATC"),
                Valid(value_link=cls.DEVICES__BREEDER_CLS.ATC.connect, name="ATC.connect()"),
            ],
        )
        return result_chain

    def run__wrapped(self) -> TYPE__RESULT_W_NORETURN:
        return True
        time.sleep(0.1)
        result_chain = ValidChains(
            [
                Valid(value_link=self.DEVICES__BREEDER_INST.DUT.VALUE, name="DUT.VALUE"),
                Valid(value_link=self.DEVICES__BREEDER_INST.ATC.address__validate),
            ],
        )
        return result_chain


# =====================================================================================================================
